# Copyright cocotb contributors
# Licensed under the Revised BSD License, see LICENSE for details.
# SPDX-License-Identifier: BSD-3-Clause
from functools import lru_cache
from typing import (
    Dict,
    Union,
)

from cocotb._py_compat import Self, TypeAlias
from cocotb.types._resolve import RESOLVE_X, ResolverLiteral, get_str_resolver

LogicLiteralT: TypeAlias = Union[str, int, bool]
LogicConstructibleT: TypeAlias = Union[LogicLiteralT, "Logic"]


_U = 0
_X = 1
_0 = 2
_1 = 3
_Z = 4
_W = 5
_L = 6
_H = 7
_D = 8

_literal_repr: Dict[LogicLiteralT, int] = {
    # unassigned
    "U": _U,
    "u": _U,
    # unknown
    "X": _X,
    "x": _X,
    # 0
    0: _0,  # Also `False`
    "0": _0,
    # 1
    1: _1,  # Also `True`
    "1": _1,
    # high impedance
    "Z": _Z,
    "z": _Z,
    # weak unknown
    "W": _W,
    "w": _W,
    # weak 0
    "L": _L,
    "l": _L,
    # weak 1
    "H": _H,
    "h": _H,
    # don't care
    "-": _D,
}


class Logic:
    r"""9-state digital signal value type.

    This type is modeled after VHDL's ``std_ulogic`` type.
    It can represent the values (``U``, ``X``, ``0``, ``1``, ``Z``, ``W``, ``L``, ``H``, ``-``).
    (System)Verilog's 4-state ``logic`` type is a subset which only utilizes the ``X``, ``0``, ``1``, and ``Z`` values.

    :class:`!Logic` can be converted to and from :class:`int`, :class:`str`, :class:`bool` and :class:`Bit`.
    String literals include ``"U"``, ``"X"``, ``"0"``, ``"1"``, ``"Z"``, ``"W"``, ``"L"``, ``"H"``, ``"-"``, and their lowercase values.

    .. code-block:: pycon3

        >>> Logic("X")
        Logic('X')
        >>> Logic(True)
        Logic('1')
        >>> Logic(1)
        Logic('1')

        >>> str(Logic("Z"))
        'Z'
        >>> bool(Logic(0))
        False
        >>> int(Logic(1))
        1

    .. note::

        The :class:`int` and :class:`bool` conversions will raise :exc:`ValueError` if the value is not ``0``, ``1``, ``L``, or ``H``.

    :class:`Logic` supports the common logic operations ``&``, ``|``, ``^``, and ``~``.

    .. code-block:: pycon3

        >>> def full_adder(a: Logic, b: Logic, carry: Logic) -> Tuple[Logic, Logic]:
        ...     res = a ^ b ^ carry
        ...     carry_out = (a & b) | (b & carry) | (a & carry)
        ...     return res, carry_out

        >>> full_adder(a=Logic("0"), b=Logic("1"), carry=Logic("1"))
        (Logic('0'), Logic('1'))

    Args:
        value: value to construct into a :class:`!Logic`.

    Raises:
        ValueError: If the value if of the correct type, but cannot be constructed into a :class:`!Logic`.
        TypeError: If the value is of a type that can't be constructed into a :class:`!Logic`.
    """

    _values = {_U, _X, _0, _1, _Z, _W, _L, _H, _D}

    _repr: int

    __slots__ = ("_repr",)

    @classmethod
    @lru_cache(maxsize=None)
    def _singleton(cls, _repr: int) -> Self:
        """Return the Logic object associated with the repr, enforcing singleton."""
        self = object.__new__(cls)
        self._repr = _repr
        return self

    def __new__(
        cls,
        value: LogicConstructibleT,
    ) -> Self:
        if isinstance(value, Logic):
            _repr = value._repr
        elif isinstance(value, (str, int)):
            try:
                _repr = _literal_repr[value]
            except KeyError:
                raise ValueError(
                    f"{value!r} is not convertible to {cls.__qualname__}"
                ) from None
        else:
            raise TypeError(
                f"Expected str, bool, or int, not {type(value).__qualname__}"
            )

        if _repr not in cls._values:
            raise ValueError(f"{value!r} is not a valid {cls.__qualname__}")

        return cls._singleton(_repr)

    def __and__(self, other: Self) -> Self:
        if not isinstance(other, type(self)):
            return NotImplemented
        return type(self)(
            (
                # -----------------------------------------------------
                # U    X    0    1    Z    W    L    H    -       |   |
                # -----------------------------------------------------
                ("U", "U", "0", "U", "U", "U", "0", "U", "U"),  # | U |
                ("U", "X", "0", "X", "X", "X", "0", "X", "X"),  # | X |
                ("0", "0", "0", "0", "0", "0", "0", "0", "0"),  # | 0 |
                ("U", "X", "0", "1", "X", "X", "0", "1", "X"),  # | 1 |
                ("U", "X", "0", "X", "X", "X", "0", "X", "X"),  # | Z |
                ("U", "X", "0", "X", "X", "X", "0", "X", "X"),  # | W |
                ("0", "0", "0", "0", "0", "0", "0", "0", "0"),  # | L |
                ("U", "X", "0", "1", "X", "X", "0", "1", "X"),  # | H |
                ("U", "X", "0", "X", "X", "X", "0", "X", "X"),  # | - |
            )[self._repr][other._repr]
        )

    def __rand__(self, other: Self) -> Self:
        return self & other

    def __or__(self, other: Self) -> Self:
        if not isinstance(other, type(self)):
            return NotImplemented
        return type(self)(
            (
                # -----------------------------------------------------
                # U    X    0    1    Z    W    L    H    -       |   |
                # -----------------------------------------------------
                ("U", "U", "U", "1", "U", "U", "U", "1", "U"),  # | U |
                ("U", "X", "X", "1", "X", "X", "X", "1", "X"),  # | X |
                ("U", "X", "0", "1", "X", "X", "0", "1", "X"),  # | 0 |
                ("1", "1", "1", "1", "1", "1", "1", "1", "1"),  # | 1 |
                ("U", "X", "X", "1", "X", "X", "X", "1", "X"),  # | Z |
                ("U", "X", "X", "1", "X", "X", "X", "1", "X"),  # | W |
                ("U", "X", "0", "1", "X", "X", "0", "1", "X"),  # | L |
                ("1", "1", "1", "1", "1", "1", "1", "1", "1"),  # | H |
                ("U", "X", "X", "1", "X", "X", "X", "1", "X"),  # | - |
            )[self._repr][other._repr]
        )

    def __ror__(self, other: Self) -> Self:
        return self | other

    def __xor__(self, other: Self) -> Self:
        if not isinstance(other, type(self)):
            return NotImplemented
        return type(self)(
            (
                # -----------------------------------------------------
                # U    X    0    1    Z    W    L    H    -       |   |
                # -----------------------------------------------------
                ("U", "U", "U", "U", "U", "U", "U", "U", "U"),  # | U |
                ("U", "X", "X", "X", "X", "X", "X", "X", "X"),  # | X |
                ("U", "X", "0", "1", "X", "X", "0", "1", "X"),  # | 0 |
                ("U", "X", "1", "0", "X", "X", "1", "0", "X"),  # | 1 |
                ("U", "X", "X", "X", "X", "X", "X", "X", "X"),  # | Z |
                ("U", "X", "X", "X", "X", "X", "X", "X", "X"),  # | W |
                ("U", "X", "0", "1", "X", "X", "0", "1", "X"),  # | L |
                ("U", "X", "1", "0", "X", "X", "1", "0", "X"),  # | H |
                ("U", "X", "X", "X", "X", "X", "X", "X", "X"),  # | - |
            )[self._repr][other._repr]
        )

    def __rxor__(self, other: Self) -> Self:
        return self ^ other

    def __invert__(self) -> Self:
        return type(self)(("U", "X", "1", "0", "X", "X", "1", "0", "X")[self._repr])

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Logic):
            return self._repr == other._repr
        elif isinstance(other, (int, str, bool)):
            try:
                other = Logic(other)
            except ValueError:
                return False
            return self == other
        else:
            return NotImplemented

    __hash__: None  # type: ignore[assignment]

    def __repr__(self) -> str:
        return f"{type(self).__qualname__}({str(self)!r})"

    def __str__(self) -> str:
        return ("U", "X", "0", "1", "Z", "W", "L", "H", "-")[self._repr]

    if RESOLVE_X is None:

        def __bool__(self) -> bool:
            if self._repr in (_0, _L):
                return False
            elif self._repr in (_1, _H):
                return True
            raise ValueError(f"Cannot convert {self!r} to bool")

        def __int__(self) -> int:
            if self._repr in (_0, _L):
                return 0
            elif self._repr in (_1, _H):
                return 1
            raise ValueError(f"Cannot convert {self!r} to int")

    else:

        def __bool__(self) -> bool:
            return self._repr in (_1, _H)

        def __int__(self) -> int:
            s = str(self)
            s = RESOLVE_X(s)
            return int(s, 2)

    def __index__(self) -> int:
        return int(self)

    def resolve(self, resolver: ResolverLiteral) -> Self:
        """Resolve non-``0``/``1`` values to ``0``/``1``.

        The possible values of the *resolver* argument are:

        * ``"weak"``:
            Weak values are resolved to their strong-valued equivalents.

        * ``"zeros"``:
            ``L`` and ``H`` are resolved to ``0`` and ``1``, respectively.
            Remaining non-``0``/``1`` values are resolved to ``0``.

        * ``"ones"``:
            ``L`` and ``H`` are resolved to ``0`` and ``1``, respectively.
            Remaining non-``0``/``1`` values are resolved to ``1``.

        * ``"random"``:
            ``L`` and ``H`` are resolved to ``0`` and ``1``, respectively.
            Remaining non-``0``/``1`` values are randomly resolved to either ``0`` or ``1``.

        Args:
            resolver: How to resolve non-``0``/``1`` values. See possible values above.

        Returns:
            The resolved Logic.

        Raises:
            ValueError: Invalid *resolver* value.
            TypeError: Unsupported *value* type.
        """
        return type(self)(get_str_resolver(resolver)(str(self)))

    def __len__(self) -> int:
        return 1

    @property
    def is_resolvable(self) -> bool:
        """``True`` if value is ``0``, ``1``, ``L``, ``H``.

        .. versionadded:: 2.0
        """
        return (False, False, True, True, False, False, True, True, False)[self._repr]

    def __copy__(self) -> "Logic":
        return self

    def __deepcopy__(self, memo: Dict[int, object]) -> "Logic":
        return self


class Bit(Logic):
    """2-state digital signal value type.

    This is modeled after (System)Verilog's and VHDL's ``bit`` type.
    It can represent only the values ``0`` and ``1``.
    It can be converted to and from :class:`int`, :class:`str`, :class:`bool`, or :class:`Logic` values.

    As a subtype of :class:`!Logic`, it supports all of the same operations
    and can be used in operations interchangeably with :class:`!Logic`.

    Args:
        value: value to construct into a :class:`!Bit`.

    Raises:
        ValueError: If the value if of the correct type, but cannot be constructed into a :class:`!Bit`.
        TypeError: If the value is of a type that can't be constructed into a :class:`!Bit`.
    """

    _values = {_0, _1}



# Copyright cocotb contributors
# Licensed under the Revised BSD License, see LICENSE for details.
# SPDX-License-Identifier: BSD-3-Clause
import copy
import warnings
from typing import Any, Dict, Iterable, Iterator, List, TypeVar, Union, cast, overload

from cocotb.types._abstract_array import AbstractMutableArray
from cocotb.types._indexing import IndexingChangedWarning
from cocotb.types._range import Range

T = TypeVar("T")


class Array(AbstractMutableArray[T]):
    r"""Fixed-size, arbitrarily-indexed, homogeneous collection type.

    Arrays are similar to, but different from Python :class:`list`\ s.
    An array can store values of any type or values of multiple types at a time, just like a :class:`!list`.
    Array constructor values can be any Iterable.

    .. code-block:: pycon3

        >>> Array([1, False, "example", None])
        Array([1, False, 'example', None], Range(0, 'to', 3))

        >>> Array("Hello!")
        Array(['H', 'e', 'l', 'l', 'o', '!'], Range(0, 'to', 5))

    Unlike :class:`!list`\ s, an array's size cannot change.
    This means they do not support methods like :meth:`~list.append` or :meth:`~list.insert`.

    The indexes of an Array can start or end at any integer value, they are not limited to 0-based indexing.
    Indexing schemes are selected using :class:`Range` objects.
    If *range* is not given, the range ``Range(0, "to", len(value)-1)`` is used.
    If an :class:`int` is passed for *range*, the range ``Range(0, "to", range-1)`` is used.

    .. code-block:: pycon3

        >>> a = Array([0, 1, 2, 3], Range(-1, "downto", -4))
        >>> a[-1]
        0
        >>> a[-4]
        3

    Arrays can also have 0 length using "null" :class:`Range`\ s.

    .. code-block:: pycon3

        >>> Array([], Range(1, "to", 0))  # 1 to 0 is a null Range
        Array([], Range(1, 'to', 0))

    Indexing and slicing is very similar to :class:`!list`\ s, but it uses the indexing scheme specified with the *range*.
    Unlike :class:`!list`, slicing uses an inclusive right bound, which is common in HDLs.
    But like :class:`!list`, if a start or stop index is not specified in the slice, it is inferred as the start or end of the array.
    Slicing an Array returns a new :class:`~cocotb.types.Array` object, whose bounds are the slice indexes.

    .. code-block:: pycon3

        >>> a = Array("1234abcd")
        >>> a[7]
        'd'
        >>> a[2:5]
        Array(['3', '4', 'a', 'b'], Range(2, 'to', 5))
        >>> a[2:5] = reversed(a[2:5])
        >>> "".join(a)
        '12ba43cd'

        >>> b = Array("1234", Range(0, -3))
        >>> b[-2]
        '3'
        >>> b[-1:]
        Array(['2', '3', '4'], Range(-1, 'downto', -3))
        >>> b[:] = reversed(b)
        >>> b
        Array(['4', '3', '2', '1'], Range(0, 'downto', -3))

    .. warning::
        Arrays behave differently in certain situations than Python's built-in sequence types (:class:`list`, :class:`tuple`, etc.).

        - Arrays are not necessarily 0-based and slices use inclusive right bounds,
          so many functions that work on Python sequences by index (like :mod:`bisect`) may not work on arrays.
        - Slice indexes must be specified in the same direction as the array and do not support specifying a "step".
        - When setting a slice, the new value must be an iterable of the same size as the slice.
        - Negative indexes are *not* treated as an offset from the end of the array, but are treated literally.

    Arrays are equal to other arrays of the same length with the same values (structural equality).
    Bounds do not matter for equality.

    .. code-block:: pycon3

        >>> a = Array([1, 1, 2, 3, 5], Range(4, "downto", 0))
        >>> b = Array([1, 1, 2, 3, 5], Range(-2, "to", 2))
        >>> a == b
        True

    You can change the bounds of an array by setting the :attr:`range` to a new value.
    The new bounds must be the same length of the array.

    .. code-block:: pycon3

        >>> a = Array("1234")
        >>> a.range
        Range(0, 'to', 3)
        >>> a.range = Range(3, "downto", 0)
        >>> a.range
        Range(3, 'downto', 0)

    Arrays support many of the methods and semantics defined by :class:`collections.abc.Sequence`.

    .. code-block:: pycon3

        >>> a = Array("stuff", Range(2, "downto", -2))
        >>> len(a)
        5
        >>> "t" in a
        True
        >>> a.index("u")
        0
        >>> for c in a:
        ...     print(c)
        s
        t
        u
        f
        f

    Args:
        value: Initial value for the Array.
        range: The indexing scheme of the Array.

    Raises:
        ValueError: When argument values cannot be used to construct an Array.
        TypeError: When invalid argument types are used.
    """

    __slots__ = ("_value", "_range", "_warn_indexing")

    def __init__(
        self, value: Iterable[T], range: Union[Range, int, None] = None
    ) -> None:
        self._warn_indexing = False
        self._value = list(value)
        if range is None:
            self._range = Range(0, "to", len(self._value) - 1)
        else:
            if isinstance(range, int):
                self._range = Range(0, "to", range - 1)
            elif isinstance(range, Range):
                self._range = range
            else:
                raise TypeError(
                    f"Expected Range or int for parameter 'range', not {type(range).__qualname__}"
                )

            if len(self._value) != len(self._range):
                raise ValueError(
                    f"Value of length {len(self._value)!r} does not fit in {self._range!r}"
                )

    @classmethod
    def _from_handle(
        cls, value: List[T], range: Range, warn_indexing: bool
    ) -> "Array[T]":
        self = cls.__new__(cls)
        self._warn_indexing = warn_indexing
        self._value = value
        self._range = range
        return self

    @property
    def range(self) -> Range:
        """:class:`Range` of the indexes of the array."""
        return self._range

    @range.setter
    def range(self, new_range: Range) -> None:
        """Sets a new indexing scheme on the array, must be the same size"""
        if not isinstance(new_range, Range):
            raise TypeError("range argument must be of type 'Range'")
        if len(new_range) != len(self):
            raise ValueError(
                f"{new_range!r} not the same length as old range ({self._range!r})."
            )
        self._range = new_range

    def __iter__(self) -> Iterator[T]:
        return iter(self._value)

    def __reversed__(self) -> Iterator[T]:
        return reversed(self._value)

    def __contains__(self, item: object) -> bool:
        return item in self._value

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Array):
            return self._value == other._value
        elif isinstance(other, list):
            return self._value == other
        elif isinstance(other, tuple):
            return tuple(self._value) == other
        else:
            return NotImplemented

    @overload
    def __getitem__(self, item: int) -> T: ...

    @overload
    def __getitem__(self, item: slice) -> "Array[T]": ...

    def __getitem__(self, item: Union[int, slice]) -> Union[T, "Array[T]"]:
        if isinstance(item, int):
            if self._warn_indexing:
                warnings.warn(
                    f"Update index {item} to {self.range[item]}",
                    IndexingChangedWarning,
                    stacklevel=2,
                )
            idx = self._translate_index(item)
            return self._value[idx]
        elif isinstance(item, slice):
            if self._warn_indexing:
                start = item.start if item.start is not None else 0
                stop = item.stop if item.stop is not None else len(self) - 1
                warnings.warn(
                    f"Update slice {start}:{stop} to {self.range[start]}:{self.range[stop]}",
                    IndexingChangedWarning,
                    stacklevel=2,
                )
            start = item.start if item.start is not None else self.left
            stop = item.stop if item.stop is not None else self.right
            if item.step is not None:
                raise IndexError("do not specify step")
            start_i = self._translate_index(start)
            stop_i = self._translate_index(stop)
            if start_i > stop_i:
                raise IndexError(
                    f"slice [{start}:{stop}] direction does not match array direction [{self.left}:{self.right}]"
                )
            value = self._value[start_i : stop_i + 1]
            range = Range(start, self.direction, stop)
            return Array(value=value, range=range)
        raise TypeError(f"indexes must be ints or slices, not {type(item).__name__}")

    @overload
    def __setitem__(self, item: int, value: T) -> None: ...

    @overload
    def __setitem__(self, item: slice, value: Iterable[T]) -> None: ...

    def __setitem__(
        self, item: Union[int, slice], value: Union[T, Iterable[T]]
    ) -> None:
        if isinstance(item, int):
            idx = self._translate_index(item)
            self._value[idx] = cast("T", value)
        elif isinstance(item, slice):
            start = item.start if item.start is not None else self.left
            stop = item.stop if item.stop is not None else self.right
            if item.step is not None:
                raise IndexError("do not specify step")
            start_i = self._translate_index(start)
            stop_i = self._translate_index(stop)
            if start_i > stop_i:
                raise IndexError(
                    f"slice [{start}:{stop}] direction does not match array direction [{self.left}:{self.right}]"
                )
            value = list(cast("Iterable[T]", value))
            if len(value) != (stop_i - start_i + 1):
                raise ValueError(
                    f"value of length {len(value)!r} will not fit in slice [{start}:{stop}]"
                )
            self._value[start_i : stop_i + 1] = value
        else:
            raise TypeError(
                f"indexes must be ints or slices, not {type(item).__name__}"
            )

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self._value!r}, {self._range!r})"

    def _translate_index(self, item: int) -> int:
        try:
            return self._range.index(item)
        except ValueError:
            raise IndexError(f"index {item} out of range") from None

    def __copy__(self) -> "Array":
        return Array(self._value, self._range)

    def __deepcopy__(self, memo: Dict[int, Any]) -> "Array":
        res = Array.__new__(Array)
        res._value = copy.deepcopy(self._value, memo=memo)
        res._range = copy.deepcopy(self._range, memo=memo)
        res._warn_indexing = self._warn_indexing
        return res



# Copyright cocotb contributors
# Licensed under the Revised BSD License, see LICENSE for details.
# SPDX-License-Identifier: BSD-3-Clause
from ._abstract_array import AbstractArray, AbstractMutableArray
from ._array import Array
from ._indexing import IndexingChangedWarning
from ._logic import Bit, Logic
from ._logic_array import LogicArray
from ._range import Range

# isort: split
# These are imports for doctests in the submodules. Since we fix up the `__module__`
# attribute, `--doctest-modules` thinks this is the module the types were defined in
# and will evaluate this module first before running tests.
from typing import Tuple  # noqa: F401

__all__ = (
    "AbstractArray",
    "AbstractMutableArray",
    "Array",
    "Bit",
    "IndexingChangedWarning",
    "Logic",
    "LogicArray",
    "Range",
)

# Set __module__ on re-exports
for name in __all__:
    obj = globals()[name]
    obj.__module__ = __name__



# Copyright cocotb contributors
# Copyright (c) 2013 Potential Ventures Ltd
# Copyright (c) 2013 SolarFlare Communications Inc
# Licensed under the Revised BSD License, see LICENSE for details.
# SPDX-License-Identifier: BSD-3-Clause

"""Tools for dealing with simulated time."""

import warnings
from decimal import Decimal
from fractions import Fraction
from functools import lru_cache
from math import ceil, floor
from typing import Union, cast, overload

from cocotb import simulator
from cocotb._py_compat import Literal, TypeAlias
from cocotb._typing import RoundMode, TimeUnit

__all__ = (
    "convert",
    "get_sim_time",
    "time_precision",
)


Steps: TypeAlias = Literal["step"]
TimeUnitWithoutSteps: TypeAlias = Literal["fs", "ps", "ns", "us", "ms", "sec"]


@overload
def convert(
    value: Union[float, Fraction, Decimal],
    unit: TimeUnit,
    *,
    to: Steps,
    round_mode: RoundMode = "error",
) -> int: ...


@overload
def convert(
    value: Union[float, Fraction, Decimal],
    unit: TimeUnit,
    *,
    to: TimeUnitWithoutSteps,
    round_mode: RoundMode = "error",
) -> float: ...


def convert(
    value: Union[float, Decimal, Fraction],
    unit: TimeUnit,
    *,
    to: TimeUnit,
    round_mode: RoundMode = "error",
) -> float:
    """Convert time values from one unit to another unit.

    Args:
        value: The time value.

        unit: The unit of *value* (one of ``'step'``, ``'fs'``, ``'ps'``, ``'ns'``, ``'us'``, ``'ms'``, ``'sec'``).

        to: The unit to convert *value* to (one of ``'step'``, ``'fs'``, ``'ps'``, ``'ns'``, ``'us'``, ``'ms'``, ``'sec'``).

        round_mode:
            How to handle non-integral step values (one of ``'error'``, ``'round'``, ``'ceil'``, ``'floor'``).

            When *round_mode* is ``"error"``, a :exc:`ValueError` is thrown if the value cannot
            be accurately represented in terms of simulator time steps.
            When *round_mode* is ``"round"``, ``"ceil"``, or ``"floor"``, the corresponding
            rounding function from the standard library will be used to round to a simulator
            time step.

    Returns:
        The value scaled by the difference in units.

    .. versionadded:: 2.0
    """
    if unit == "step":
        steps = cast("int", value)
    else:
        steps = _get_sim_steps(value, unit, round_mode=round_mode)
    if to == "step":
        return steps
    else:
        return _get_time_from_sim_steps(steps, to)


@overload
def get_sim_time(unit: Steps = "step", *, units: None = None) -> int: ...


@overload
def get_sim_time(unit: TimeUnitWithoutSteps, *, units: None = None) -> float: ...


def get_sim_time(unit: TimeUnit = "step", *, units: None = None) -> float:
    """Retrieve the simulation time from the simulator.

    Args:
        unit: String specifying the unit of the result
            (one of ``'step'``, ``'fs'``, ``'ps'``, ``'ns'``, ``'us'``, ``'ms'``, ``'sec'``).
            ``'step'`` will return the raw simulation time.

            .. versionchanged:: 2.0
                Passing ``None`` as the *unit* argument was removed, use ``'step'`` instead.

            .. versionchanged:: 2.0
                Renamed from ``units``.

    Raises:
        ValueError: If *unit* is not a valid unit.

    Returns:
        The simulation time in the specified unit.

    .. versionchanged:: 1.6
        Support ``'step'`` as the the *unit* argument to mean "simulator time step".

    .. versionchanged:: 2.0
        Moved from :mod:`cocotb.utils` to :mod:`cocotb.simtime`.
    """
    if units is not None:
        warnings.warn(
            "The 'units' argument has been renamed to 'unit'.",
            DeprecationWarning,
            stacklevel=2,
        )
        unit = units
    timeh, timel = simulator.get_sim_time()
    steps = timeh << 32 | timel
    return _get_time_from_sim_steps(steps, unit) if unit != "step" else steps


@overload
def _ldexp10(frac: float, exp: int) -> float: ...


@overload
def _ldexp10(frac: Fraction, exp: int) -> Fraction: ...


@overload
def _ldexp10(frac: Decimal, exp: int) -> Decimal: ...


def _ldexp10(
    frac: Union[float, Fraction, Decimal], exp: int
) -> Union[float, Fraction, Decimal]:
    """Like :func:`math.ldexp`, but base 10."""
    # using * or / separately prevents rounding errors if `frac` is a
    # high-precision type
    if exp > 0:
        return frac * (10**exp)
    else:
        return frac / (10**-exp)


def _get_time_from_sim_steps(
    steps: int,
    unit: TimeUnit,
) -> float:
    if unit == "step":
        return steps
    return _ldexp10(steps, time_precision - _get_log_time_scale(unit))


def _get_sim_steps(
    time: Union[float, Fraction, Decimal],
    unit: TimeUnit = "step",
    *,
    round_mode: RoundMode = "error",
) -> int:
    result: Union[float, Fraction, Decimal]
    if unit != "step":
        result = _ldexp10(time, _get_log_time_scale(unit) - time_precision)
    else:
        result = time

    if round_mode == "error":
        result_rounded = floor(result)
        if result_rounded != result:
            raise ValueError(
                f"Unable to accurately represent {time}({unit}) with the simulator precision of 1e{time_precision}"
            )
    elif round_mode == "ceil":
        result_rounded = ceil(result)
    elif round_mode == "round":
        result_rounded = round(result)
    elif round_mode == "floor":
        result_rounded = floor(result)
    else:
        raise ValueError(f"Invalid round_mode specifier: {round_mode}")

    return result_rounded


@lru_cache(maxsize=None)
def _get_log_time_scale(unit: TimeUnitWithoutSteps) -> int:
    """Retrieve the ``log10()`` of the scale factor for a given time unit.

    Args:
        unit: String specifying the unit
            (one of ``'fs'``, ``'ps'``, ``'ns'``, ``'us'``, ``'ms'``, ``'sec'``).

            .. versionchanged:: 2.0
                Renamed from ``units``.

    Raises:
        ValueError: If *unit* is not a valid unit.

    Returns:
        The ``log10()`` of the scale factor for the time unit.
    """
    scale = {"fs": -15, "ps": -12, "ns": -9, "us": -6, "ms": -3, "sec": 0}

    unit_lwr = unit.lower()
    if unit_lwr not in scale:
        raise ValueError(f"Invalid unit ({unit}) provided")
    else:
        return scale[unit_lwr]


time_precision: int = _get_log_time_scale("fs")
"""The precision of time in the current simulation.

The value is seconds in powers of tens,
i.e. ``-15`` is ``10**-15`` seconds or 1 femtosecond.

.. versionadded:: 2.0
"""


def _init() -> None:
    global time_precision
    time_precision = simulator.get_precision()



# Copyright cocotb contributors
# Licensed under the Revised BSD License, see LICENSE for details.
# SPDX-License-Identifier: BSD-3-Clause

# type: ignore  # distutils is untyped, so there's little reason to check this file

"""
Classes to compare simulation versions.

These are for cocotb-internal use only.

.. warning::
    These classes silently allow comparing versions of different simulators.
"""

from cocotb_tools._vendor.distutils_version import LooseVersion


class ActivehdlVersion(LooseVersion):
    """Version numbering class for Aldec Active-HDL.

    NOTE: unsupported versions exist, e.g.
    ActivehdlVersion("10.5a.12.6914") > ActivehdlVersion("10.5.216.6767")
    """


class CvcVersion(LooseVersion):
    """Version numbering class for Tachyon DA CVC.

    Example:
        >>> CvcVersion(
        ...     "OSS_CVC_7.00b-x86_64-rhel6x of 07/07/14 (Linux-elf)"
        ... ) > CvcVersion("OSS_CVC_7.00a-x86_64-rhel6x of 07/07/14 (Linux-elf)")
        True
    """


class GhdlVersion(LooseVersion):
    """Version numbering class for GHDL."""


class IcarusVersion(LooseVersion):
    """Version numbering class for Icarus Verilog.

    Example:
        >>> IcarusVersion("11.0 (devel)") > IcarusVersion("10.3 (stable)")
        True
        >>> IcarusVersion("10.3 (stable)") <= IcarusVersion("10.3 (stable)")
        True
    """


class ModelsimVersion(LooseVersion):
    """Version numbering class for Mentor ModelSim."""


class QuestaVersion(LooseVersion):
    """Version numbering class for Mentor Questa.

    Example:
        >>> QuestaVersion("10.7c 2018.08") > QuestaVersion("10.7b 2018.06")
        True
        >>> QuestaVersion("2020.1 2020.01") > QuestaVersion("10.7c 2018.08")
        True
        >>> QuestaVersion("2020.1 2020.01") == QuestaVersion("2020.1")
        True
        >>> QuestaVersion("2023.1_2 2023.03") > QuestaVersion("2023.1_1")
        True
    """

    def parse(self, vstring):
        # A Questa version string, as returned by the simulator, consists of two
        # space-separated parts. The first part is the actual version number,
        # the second part seems to be the year and month of the initial release.
        # We only need the first part, which is also used in public
        # communication by Siemens.
        try:
            first_component = vstring.split(" ", 1)[0]
        except IndexError:
            first_component = vstring

        super().parse(first_component)


class RivieraVersion(LooseVersion):
    """Version numbering class for Aldec Riviera-PRO.

    Example:
        >>> RivieraVersion("2019.10.138.7537") == RivieraVersion("2019.10.138.7537")
        True
    """


class VcsVersion(LooseVersion):
    """Version numbering class for Synopsys VCS.

    Example:
        >>> VcsVersion("Q-2020.03-1_Full64") > VcsVersion("K-2015.09_Full64")
        True
    """


class VerilatorVersion(LooseVersion):
    """Version numbering class for Verilator.

    Example:
        >>> VerilatorVersion("4.032 2020-04-04") > VerilatorVersion("4.031 devel")
        True
    """


class XceliumVersion(LooseVersion):
    """Version numbering class for Cadence Xcelium.

    Example:
        >>> XceliumVersion("20.06-g183") > XceliumVersion("20.03-s002")
        True
        >>> XceliumVersion("20.07-e501") > XceliumVersion("20.06-g183")
        True
    """


class IusVersion(XceliumVersion):  # inherit everything from Xcelium
    """Version numbering class for Cadence IUS.

    Example:
        >>> IusVersion("15.20-s050") > IusVersion("15.20-s049")
        True
    """


class NvcVersion(LooseVersion):
    """Version numbering class for NVC."""

    @classmethod
    def from_commandline(cls, cmdline):
        firstline = cmdline.split("\n")[0]
        sim, version, *version_extra = firstline.strip().split(" ")
        assert sim == "nvc"
        return cls(version)



# Copyright cocotb contributors
# Licensed under the Revised BSD License, see LICENSE for details.
# SPDX-License-Identifier: BSD-3-Clause
import copy
from functools import lru_cache
from typing import Any, Dict, Iterator, Sequence, Union, overload

from cocotb._utils import cached_method


class Range(Sequence[int]):
    r"""
    Variant of :class:`range` with inclusive right bound.

    In Python, :class:`range` and :class:`slice` have a non-inclusive right bound.
    In both Verilog and VHDL, ranges and arrays have an inclusive right bound.
    This type mimics Python's :class:`range` type, but implements HDL-like inclusive right bounds,
    using the names :attr:`left` and :attr:`right` as replacements for ``start`` and ``stop`` to
    match VHDL.
    Range directionality can be specified using ``'to'`` or ``'downto'`` between the
    left and right bounds.
    Not specifying directionality will cause the directionality to be inferred.

    .. code-block:: pycon3

        >>> r = Range(-2, 3)
        >>> r.left, r.right, len(r)
        (-2, 3, 6)

        >>> s = Range(8, "downto", 1)
        >>> s.left, s.right, len(s)
        (8, 1, 8)

    :meth:`from_range` and :meth:`to_range` can be used to convert from and to :class:`range`.

    .. code-block:: pycon3

        >>> r = Range(-2, 3)
        >>> r.to_range()
        range(-2, 4)

    :class:`Range` supports "null" ranges as seen in VHDL.
    "null" ranges occur when a left bound cannot reach a right bound with the given direction.
    They have a length of ``0``, but the :attr:`left`, :attr:`right`, and :attr:`direction` values remain as given.

    .. code-block:: pycon3

        >>> r = Range(1, "to", 0)  # no way to count from 1 'to' 0
        >>> r.left, r.direction, r.right
        (1, 'to', 0)
        >>> len(r)
        0

    .. note::
        This is only possible when specifying the direction.

    Ranges also support all the features of :class:`range` including, but not limited to:

    - ``value in range`` to see if a value is in the range,
    - ``range.index(value)`` to see what position in the range the value is,

    The typical use case of this type is in conjunction with :class:`~cocotb.types.Array`.

    Args:
        left: Leftmost bound of range.
        direction: ``'to'`` if values are ascending, ``'downto'`` if descending.
        right: Rightmost bound of range (inclusive).
    """

    @overload
    def __init__(self, left: int, direction: int) -> None: ...

    @overload
    def __init__(self, left: int, direction: str, right: int) -> None: ...

    @overload
    def __init__(self, left: int, *, right: int) -> None: ...

    def __init__(
        self,
        left: int,
        direction: Union[int, str, None] = None,
        right: Union[int, None] = None,
    ) -> None:
        start = left
        stop: int
        step: int
        if isinstance(direction, int) and right is None:
            step = _guess_step(left, direction)
            stop = direction + step
        elif isinstance(direction, str) and isinstance(right, int):
            step = _direction_to_step(direction)
            stop = right + step
        elif direction is None and isinstance(right, int):
            step = _guess_step(left, right)
            stop = right + step
        else:
            raise TypeError("invalid arguments")
        self._range = range(start, stop, step)

    @classmethod
    def from_range(cls, range: range) -> "Range":
        """Convert :class:`range` to :class:`Range`."""
        return cls(
            left=range.start,
            direction=_step_to_direction(range.step),
            right=(range.stop - range.step),
        )

    def to_range(self) -> range:
        """Convert Range to :class:`range`."""
        return self._range

    @property
    def left(self) -> int:
        """Leftmost value in a Range."""
        return self._range.start

    @property
    def direction(self) -> str:
        """``'to'`` if Range is ascending, ``'downto'`` otherwise."""
        return _step_to_direction(self._range.step)

    @property
    def right(self) -> int:
        """Rightmost value in a Range."""
        return self._range.stop - self._range.step

    def __len__(self) -> int:
        return len(self._range)

    @overload
    def __getitem__(self, item: int) -> int: ...

    @overload
    def __getitem__(self, item: slice) -> "Range": ...

    def __getitem__(self, item: Union[int, slice]) -> Union[int, "Range"]:
        if isinstance(item, int):
            return self._range[item]
        elif isinstance(item, slice):
            return type(self).from_range(self._range[item])
        raise TypeError(
            f"indices must be integers or slices, not {type(item).__name__}"
        )

    def __contains__(self, item: object) -> bool:
        return item in self._range

    def __iter__(self) -> Iterator[int]:
        return iter(self._range)

    def __reversed__(self) -> Iterator[int]:
        return reversed(self._range)

    def __eq__(self, other: object) -> bool:
        if isinstance(other, type(self)):
            return self._range == other._range
        return NotImplemented  # must not be in a type narrowing context to be ignored properly

    def __hash__(self) -> int:
        return hash(self._range)

    def __repr__(self) -> str:
        return f"{type(self).__qualname__}({self.left!r}, {self.direction!r}, {self.right!r})"

    index = cached_method(Sequence.index)

    def __copy__(self) -> "Range":
        return Range.from_range(self._range)

    def __deepcopy__(self, memo: Dict[int, Any]) -> "Range":
        return Range.from_range(copy.deepcopy(self._range, memo=memo))


def _guess_step(left: int, right: int) -> int:
    if left <= right:
        return 1
    return -1


@lru_cache(maxsize=None)
def _direction_to_step(direction: str) -> int:
    direction = direction.lower()
    if direction == "to":
        return 1
    elif direction == "downto":
        return -1
    raise ValueError("direction must be 'to' or 'downto'")


def _step_to_direction(step: int) -> str:
    if step == 1:
        return "to"
    elif step == -1:
        return "downto"
    raise ValueError("step must be 1 or -1")



# Copyright cocotb contributors
# Copyright (c) 2013 Potential Ventures Ltd
# Copyright (c) 2013 SolarFlare Communications Inc
# Licensed under the Revised BSD License, see LICENSE for details.
# SPDX-License-Identifier: BSD-3-Clause
import ast
import logging
import os
import random
import sys
import time
import warnings
from pathlib import Path
from types import SimpleNamespace
from typing import Callable, List, cast

import cocotb
import cocotb._profiling
import cocotb.handle
import cocotb.logging
import cocotb.simtime
import cocotb.simulator
from cocotb._scheduler import Scheduler
from cocotb.regression import RegressionManager, RegressionMode

log: logging.Logger


def _setup_logging() -> None:
    cocotb.log = logging.getLogger("test")
    cocotb.log.setLevel(logging.INFO)

    global log
    log = logging.getLogger("cocotb")


_shutdown_callbacks: List[Callable[[], None]] = []
"""List of callbacks to be called when cocotb shuts down."""


def _register_shutdown_callback(cb: Callable[[], None]) -> None:
    """Register a callback to be called when cocotb shuts down."""
    _shutdown_callbacks.append(cb)


def _shutdown_testbench() -> None:
    """Call all registered shutdown callbacks."""
    while _shutdown_callbacks:
        cb = _shutdown_callbacks.pop(0)
        cb()


def init_package_from_simulation(argv: List[str]) -> None:
    """Initialize the cocotb package from a simulation context."""

    # register a callback to be called if the simulation fails
    cocotb.simulator.set_sim_event_callback(_sim_event)

    cocotb.is_simulation = True

    cocotb.argv = argv

    # sys.path normally includes "" (the current directory), but does not appear to when python is embedded.
    # Add it back because users expect to be able to import files in their test directory.
    sys.path.insert(0, "")

    cocotb.logging._init()
    _setup_logging()

    # From https://www.python.org/dev/peps/pep-0565/#recommended-filter-settings-for-test-runners
    # If the user doesn't want to see these, they can always change the global
    # warning settings in their test module.
    if not sys.warnoptions:
        warnings.simplefilter("default")

    cocotb.SIM_NAME = cocotb.simulator.get_simulator_product().strip()
    cocotb.SIM_VERSION = cocotb.simulator.get_simulator_version().strip()

    log.info("Running on %s version %s", cocotb.SIM_NAME, cocotb.SIM_VERSION)

    cocotb._profiling.initialize()
    _register_shutdown_callback(cocotb._profiling.finalize)

    _process_plusargs()
    _process_packages()
    _setup_random_seed()
    _setup_root_handle()
    _start_user_coverage()

    cocotb.simtime._init()

    log.info(
        "Initialized cocotb v%s from %s",
        cocotb.__version__,
        Path(__file__).parent.absolute(),
    )


def run_regression(_: object) -> None:
    """Setup and run a regression."""

    _setup_regression_manager()

    # setup global scheduler system
    cocotb._scheduler_inst = Scheduler()

    # start Regression Manager
    log.info("Running tests")
    cocotb._regression_manager.start_regression()


def _sim_event(msg: str) -> None:
    """Function that can be called externally to signal an event."""
    # We simply return here as the simulator will exit
    # so no cleanup is needed
    if hasattr(cocotb, "_regression_manager"):
        cocotb._regression_manager._fail_simulation(msg)
    else:
        log.error(msg)
        _shutdown_testbench()


def _process_plusargs() -> None:
    cocotb.plusargs = {}

    for option in cocotb.argv:
        if option.startswith("+"):
            if option.find("=") != -1:
                (name, value) = option[1:].split("=", 1)
                cocotb.plusargs[name] = value
            else:
                cocotb.plusargs[option[1:]] = True


def _process_packages() -> None:
    pkg_dict = {}

    from cocotb import simulator  # noqa: PLC0415

    pkgs = simulator.package_iterate()
    if pkgs is None:
        cocotb.packages = SimpleNamespace()
        return

    for pkg in pkgs:
        handle = cast(
            "cocotb.handle.HierarchyObject", cocotb.handle._make_sim_object(pkg)
        )
        name = handle._name

        # Icarus doesn't support named access to package objects:
        # https://github.com/steveicarus/iverilog/issues/1038
        # so we cannot lazily create handles
        if cocotb.SIM_NAME == "Icarus Verilog":
            handle._discover_all()
        pkg_dict[name] = handle

    cocotb.packages = SimpleNamespace(**pkg_dict)


def _start_user_coverage() -> None:
    coverage_envvar = os.getenv("COCOTB_USER_COVERAGE")
    if coverage_envvar is None:
        coverage_envvar = os.getenv("COVERAGE")
        if coverage_envvar is not None:
            warnings.warn(
                "COVERAGE is deprecated in favor of COCOTB_USER_COVERAGE",
                DeprecationWarning,
                stacklevel=2,
            )
    if coverage_envvar:
        try:
            import coverage  # noqa: PLC0415
        except ImportError:
            raise RuntimeError(
                "Coverage collection requested but coverage module not available. Install it using `pip install coverage`."
            ) from None
        else:
            config_filepath = os.getenv("COVERAGE_RCFILE")
            if config_filepath is None:
                # Exclude cocotb itself from coverage collection.
                log.info(
                    "Collecting coverage of user code. No coverage config file supplied via COVERAGE_RCFILE."
                )
                cocotb_package_dir = Path(__file__).parent.absolute()
                user_coverage = coverage.coverage(
                    branch=True, omit=[f"{cocotb_package_dir}/*"]
                )
            else:
                log.info(
                    "Collecting coverage of user code. Coverage config file supplied."
                )
                # Allow the config file to handle all configuration
                user_coverage = coverage.coverage(config_file=config_filepath)
            user_coverage.start()

            def stop_user_coverage() -> None:
                user_coverage.stop()
                log.debug("Writing user coverage data")
                user_coverage.save()

            _register_shutdown_callback(stop_user_coverage)


def _setup_random_seed() -> None:
    seed_envvar = os.getenv("COCOTB_RANDOM_SEED")
    if seed_envvar is None:
        seed_envvar = os.getenv("RANDOM_SEED")
        if seed_envvar is not None:
            warnings.warn(
                "RANDOM_SEED is deprecated in favor of COCOTB_RANDOM_SEED",
                DeprecationWarning,
            )
    if seed_envvar is None:
        if "ntb_random_seed" in cocotb.plusargs:
            warnings.warn(
                "Passing +ntb_random_seed will not be used to seed Python's random number generator in the future. "
                "Ensure you also set `COCOTB_RANDOM_SEED`.",
                FutureWarning,
            )
            plusarg_seed = cocotb.plusargs["ntb_random_seed"]
            if not isinstance(plusarg_seed, str):
                raise TypeError("ntb_random_seed plusarg is not a valid seed value.")
            seed = ast.literal_eval(plusarg_seed)
            if not isinstance(seed, int):
                raise TypeError("ntb_random_seed plusargs is not a valid seed value.")
            cocotb.RANDOM_SEED = seed
        elif "seed" in cocotb.plusargs:
            warnings.warn(
                "Passing +seed will not be used to seed Python's random number generator in the future. "
                "Ensure you also set `COCOTB_RANDOM_SEED`.",
                FutureWarning,
            )
            plusarg_seed = cocotb.plusargs["seed"]
            if not isinstance(plusarg_seed, str):
                raise TypeError("seed plusarg is not a valid seed value.")
            seed = ast.literal_eval(plusarg_seed)
            if not isinstance(seed, int):
                raise TypeError("seed plusargs is not a valid seed value.")
            cocotb.RANDOM_SEED = seed
        else:
            cocotb.RANDOM_SEED = int(time.time())
        log.info("Seeding Python random module with %d", cocotb.RANDOM_SEED)
    else:
        cocotb.RANDOM_SEED = ast.literal_eval(seed_envvar)
        log.info(
            "Seeding Python random module with supplied seed %d", cocotb.RANDOM_SEED
        )

    random.seed(cocotb.RANDOM_SEED)


def _setup_root_handle() -> None:
    root_name = os.getenv("COCOTB_TOPLEVEL")
    if root_name is not None:
        root_name = root_name.strip()
        if root_name == "":
            root_name = None
        elif "." in root_name:
            # Skip any library component of the toplevel
            root_name = root_name.split(".", 1)[1]

    from cocotb import simulator  # noqa: PLC0415

    handle = simulator.get_root_handle(root_name)
    if not handle:
        raise RuntimeError(f"Can not find root handle {root_name!r}")

    cocotb.top = cocotb.handle._make_sim_object(handle)


def _setup_regression_manager() -> None:
    cocotb._regression_manager = RegressionManager()

    # discover tests
    module_str = os.getenv("COCOTB_TEST_MODULES", "")
    if not module_str:
        raise RuntimeError(
            "Environment variable COCOTB_TEST_MODULES, which defines the module(s) to execute, is not defined or empty."
        )
    modules = [s.strip() for s in module_str.split(",") if s.strip()]
    cocotb._regression_manager.setup_pytest_assertion_rewriting()
    cocotb._regression_manager.discover_tests(*modules)

    # filter tests
    testcase_str = os.getenv("COCOTB_TESTCASE", "").strip()
    test_filter_str = os.getenv("COCOTB_TEST_FILTER", "").strip()
    if testcase_str and test_filter_str:
        raise RuntimeError("Specify only one of COCOTB_TESTCASE or COCOTB_TEST_FILTER")
    elif testcase_str:
        warnings.warn(
            "COCOTB_TESTCASE is deprecated in favor of COCOTB_TEST_FILTER",
            DeprecationWarning,
            stacklevel=2,
        )
        filters = [f"{s.strip()}$" for s in testcase_str.split(",") if s.strip()]
        cocotb._regression_manager.add_filters(*filters)
        cocotb._regression_manager.set_mode(RegressionMode.TESTCASE)
    elif test_filter_str:
        cocotb._regression_manager.add_filters(test_filter_str)
        cocotb._regression_manager.set_mode(RegressionMode.TESTCASE)



# Copyright cocotb contributors
# Copyright (c) 2013 Potential Ventures Ltd
# Copyright (c) 2013 SolarFlare Communications Inc
# Licensed under the Revised BSD License, see LICENSE for details.
# SPDX-License-Identifier: BSD-3-Clause

import re
import sys
import xml.etree.ElementTree as ET
from typing import Union
from xml.etree.ElementTree import Element, SubElement


# Shamelessly ripped from pytest source code.
# https://github.com/pytest-dev/pytest/blob/d036b12bb6fa09f9a8a3b690cc7336113c93fa44/src/_pytest/junitxml.py#L37C1-L61C50
def bin_xml_escape(arg: object) -> str:
    r"""Visually escape invalid XML characters.

    For example, transforms
        'hello\aworld\b'
    into
        'hello#x07world#x08'
    Note that the #xABs are *not* XML escapes - missing the ampersand &#xAB.
    The idea is to escape visually for the user rather than for XML itself.
    """

    def repl(matchobj: "re.Match[str]") -> str:
        i = ord(matchobj.group())
        if i <= 0xFF:
            return f"#x{i:02X}"
        else:
            return f"#x{i:04X}"

    # The spec range of valid chars is:
    # Char ::= #x9 | #xA | #xD | [#x20-#xD7FF] | [#xE000-#xFFFD] | [#x10000-#x10FFFF]
    # For an unknown(?) reason, we disallow #x7F (DEL) as well.
    illegal_xml_re = (
        "[^\u0009\u000a\u000d\u0020-\u007e\u0080-\ud7ff\ue000-\ufffd\u10000-\u10ffff]"
    )
    return re.sub(illegal_xml_re, repl, str(arg))


if sys.version_info < (3, 9):

    def indent(elem: Element, level: int = 0) -> None:
        i = "\n" + level * "  "
        if len(elem):
            if not elem.text or not elem.text.strip():
                elem.text = i + "  "
            if not elem.tail or not elem.tail.strip():
                elem.tail = i
            for sub_elem in elem:
                indent(sub_elem, level + 1)
            if not sub_elem.tail or not sub_elem.tail.strip():
                sub_elem.tail = i
        elif level and (not elem.tail or not elem.tail.strip()):
            elem.tail = i

else:
    from xml.etree.ElementTree import indent


class XUnitReporter:
    last_testsuite: Element
    last_testcase: Element

    def __init__(self, filename: str = "results.xml") -> None:
        self.results = Element("testsuites", name="results")
        self.filename = filename

    def add_testsuite(self, **kwargs: str) -> Element:
        self.last_testsuite = SubElement(self.results, "testsuite", kwargs)
        return self.last_testsuite

    def add_testcase(
        self, testsuite: Union[Element, None] = None, **kwargs: str
    ) -> Element:
        if testsuite is None:
            testsuite = self.last_testsuite
        self.last_testcase = SubElement(testsuite, "testcase", kwargs)
        return self.last_testcase

    def add_property(
        self, testsuite: Union[Element, None] = None, **kwargs: str
    ) -> Element:
        if testsuite is None:
            testsuite = self.last_testsuite
        self.last_property = SubElement(testsuite, "property", kwargs)
        return self.last_property

    def add_failure(self, testcase: Union[Element, None] = None, **kwargs: str) -> None:
        if testcase is None:
            testcase = self.last_testcase
        SubElement(testcase, "failure", kwargs)

    def add_skipped(self, testcase: Union[Element, None] = None, **kwargs: str) -> None:
        if testcase is None:
            testcase = self.last_testcase
        SubElement(testcase, "skipped", kwargs)

    def write(self) -> None:
        indent(self.results)
        ET.ElementTree(self.results).write(self.filename, encoding="UTF-8")



#!/usr/bin/env python
# Copyright cocotb contributors
# Copyright (c) 2013 Potential Ventures Ltd
# Copyright (c) 2013 SolarFlare Communications Inc
# Licensed under the Revised BSD License, see LICENSE for details.
# SPDX-License-Identifier: BSD-3-Clause

import sys

if sys.version_info[:3] < (3, 6, 2):  # noqa: UP036 | bug in ruff
    version_str = ".".join(sys.version_info[:3])
    msg = [
        "This version of cocotb requires at least Python 3.6.2,",
        f"you are running Python {version_str}."
        "For more information please refer to the documentation at ",
        "https://cocotb.readthedocs.io.",
    ]

    raise SystemExit("\n".join(msg))

import logging
import os
import subprocess
from io import StringIO
from os import path, walk

from setuptools import find_packages, setup

# Note: cocotb is not installed properly yet and is missing dependencies and binaries
# We can still import other files next to setup.py, as long as they're in MANIFEST.in
# The below line is necessary for PEP517 support
sys.path.append(path.dirname(__file__))
from cocotb_build_libs import build_ext, get_ext

__version__ = "2.0.1"

max_python3_minor_version = 13
if "COCOTB_IGNORE_PYTHON_REQUIRES" not in os.environ and sys.version_info >= (
    3,
    max_python3_minor_version + 1,
):
    raise RuntimeError(
        f"cocotb {__version__} only supports a maximum Python version of 3.{max_python3_minor_version}.\n"
        "You can suppress this error by defining the environment variable COCOTB_IGNORE_PYTHON_REQUIRES\n"
        "There is no guarantee this will work and no support will be provided."
    )


def read_file(fname):
    with open(path.join(path.dirname(__file__), fname), encoding="utf8") as f:
        return f.read()


def package_files(directory):
    paths = []
    for fpath, _, filenames in walk(directory):
        for filename in filenames:
            paths.append(path.join("..", "..", fpath, filename))
    return paths


version_file_path = path.join("src", "cocotb", "_version.py")
if "dev" in __version__:
    try:
        rev = subprocess.check_output(
            ["git", "rev-parse", "--short", "HEAD"], universal_newlines=True
        ).strip()
        __version__ += f"+{rev}"
    except Exception as e:
        # if this is not a git repository and _version.py already exists,
        # we are probably installing from an sdist, so use the existing _version.py
        if path.exists(version_file_path):
            exec(read_file(version_file_path))
        else:
            print(e, file=sys.stderr)
with open(version_file_path, "w") as f:
    f.write("# Package version\n")
    f.write("# Generated by setup.py -- do not modify directly\n\n")
    f.write(f'__version__ = "{__version__}"\n')


# store log from build_libs and display at the end in verbose mode
# see https://github.com/pypa/pip/issues/6634
log_stream = StringIO()
handler = logging.StreamHandler(log_stream)
log = logging.getLogger("cocotb_build_libs")
log.setLevel(logging.INFO)
log.addHandler(handler)

setup(
    name="cocotb",
    cmdclass={"build_ext": build_ext},
    version=__version__,
    description="cocotb is a coroutine based cosimulation library for writing VHDL and Verilog testbenches in Python.",
    url="https://www.cocotb.org",
    license="BSD-3-Clause",
    long_description=read_file("README.md"),
    long_description_content_type="text/markdown",
    author="Chris Higgs, Stuart Hodgson",
    maintainer="cocotb contributors",
    maintainer_email="cocotb@lists.librecores.org",
    install_requires=[
        "find_libpython",
    ],
    python_requires=">=3.6.2",
    packages=find_packages(where="src"),
    package_dir={"": "src"},
    package_data={
        "cocotb": (
            package_files("src/cocotb/share/include")
            + package_files("src/cocotb/share/def")
            + package_files("src/cocotb/share/lib/verilator")
        ),
        "cocotb_tools": (package_files("src/cocotb_tools/makefiles")),
    },
    ext_modules=get_ext(),
    entry_points={
        "console_scripts": [
            "cocotb-config=cocotb_tools.config:main",
        ]
    },
    platforms="any",
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.13",
        "Topic :: Scientific/Engineering :: Electronic Design Automation (EDA)",
        "Framework :: cocotb",
    ],
    # these appear in the sidebar on PyPI
    project_urls={
        "Bug Tracker": "https://github.com/cocotb/cocotb/issues",
        "Source Code": "https://github.com/cocotb/cocotb",
        "Documentation": "https://docs.cocotb.org",
    },
)

print(log_stream.getvalue())



# Copyright cocotb contributors
# Copyright (c) 2013 Potential Ventures Ltd
# Copyright (c) 2013 SolarFlare Communications Inc
# Licensed under the Revised BSD License, see LICENSE for details.
# SPDX-License-Identifier: BSD-3-Clause

"""Tools for dealing with simulated time."""

import warnings
from decimal import Decimal
from fractions import Fraction
from typing import Union

from cocotb._typing import RoundMode, TimeUnit
from cocotb.simtime import (
    _get_sim_steps,
    _get_time_from_sim_steps,
    get_sim_time,
)

__all__ = (
    "get_sim_steps",
    "get_sim_time",
    "get_time_from_sim_steps",
)


def get_time_from_sim_steps(
    steps: int,
    unit: Union[TimeUnit, None] = None,
    *,
    units: None = None,
) -> float:
    """Calculate simulation time in the specified *unit* from the *steps* based
    on the simulator precision.

    Args:
        steps: Number of simulation steps.
        unit: String specifying the unit of the result
            (one of ``'fs'``, ``'ps'``, ``'ns'``, ``'us'``, ``'ms'``, ``'sec'``).

            .. versionchanged:: 2.0
                Renamed from ``units``.

    Raises:
        ValueError: If *unit* is not a valid unit.

    Returns:
        The simulation time in the specified unit.
    """
    if units is not None:
        warnings.warn(
            "The 'units' argument has been renamed to 'unit'.",
            DeprecationWarning,
            stacklevel=2,
        )
        unit = units
    if unit is None:
        raise TypeError("Missing required argument 'unit'")
    return _get_time_from_sim_steps(steps, unit)


def get_sim_steps(
    time: Union[float, Fraction, Decimal],
    unit: TimeUnit = "step",
    *,
    round_mode: RoundMode = "error",
    units: None = None,
) -> int:
    """Calculates the number of simulation time steps for a given amount of *time*.

    When *round_mode* is ``"error"``, a :exc:`ValueError` is thrown if the value cannot
    be accurately represented in terms of simulator time steps.
    When *round_mode* is ``"round"``, ``"ceil"``, or ``"floor"``, the corresponding
    rounding function from the standard library will be used to round to a simulator
    time step.

    Args:
        time: The value to convert to simulation time steps.
        unit: String specifying the unit of the result
            (one of ``'step'``, ``'fs'``, ``'ps'``, ``'ns'``, ``'us'``, ``'ms'``, ``'sec'``).
            ``'step'`` means *time* is already in simulation time steps.

            .. versionchanged:: 2.0
                Renamed from ``units``.

        round_mode: String specifying how to handle time values that sit between time steps
            (one of ``'error'``, ``'round'``, ``'ceil'``, ``'floor'``).

    Returns:
        The number of simulation time steps.

    Raises:
        ValueError: if the value cannot be represented accurately in terms of simulator
            time steps when *round_mode* is ``"error"``.

    .. versionchanged:: 1.5
        Support ``'step'`` as the *unit* argument to mean "simulator time step".

    .. versionchanged:: 1.6
        Support rounding modes.
    """
    if units is not None:
        warnings.warn(
            "The 'units' argument has been renamed to 'unit'.",
            DeprecationWarning,
            stacklevel=2,
        )
        unit = units
    return _get_sim_steps(time, unit, round_mode=round_mode)



# Copyright cocotb contributors
# Copyright (c) 2013, 2018 Potential Ventures Ltd
# Copyright (c) 2013 SolarFlare Communications Inc
# Licensed under the Revised BSD License, see LICENSE for details.
# SPDX-License-Identifier: BSD-3-Clause

from cocotb._py_compat import StrEnum

_ESCAPE = "\033["


class ANSI(StrEnum):
    """ANSI escape codes for coloring output.

    The color names supported are ``[BRIGHT_]{BLACK|RED|GREEN|YELLOW|BLUE|MAGENTA|CYAN|WHITE}{_FG|_BG}``.

    Variables that end in ``_FG`` will color the character or symbol ("foreground")
    and variables that end in ``_BG`` will color the background.

    Foreground and background colors can be combined together with a ``+``.
    Setting a new foreground color will override the previous foreground, likewise with background colors.

    Use ``DEFAULT_FG`` and ``DEFAULT_BG`` to reset the coloring to the default colors for the foreground and background, respectively.
    Or use ``DEFAULT`` to reset both.
    """

    DEFAULT_FG = _ESCAPE + "39m"
    DEFAULT_BG = _ESCAPE + "49m"
    DEFAULT = DEFAULT_BG + DEFAULT_FG

    BLACK_FG = _ESCAPE + "30m"
    RED_FG = _ESCAPE + "31m"
    GREEN_FG = _ESCAPE + "32m"
    YELLOW_FG = _ESCAPE + "33m"
    BLUE_FG = _ESCAPE + "34m"
    MAGENTA_FG = _ESCAPE + "35m"
    CYAN_FG = _ESCAPE + "36m"
    WHITE_FG = _ESCAPE + "37m"

    BLACK_BG = _ESCAPE + "40m"
    RED_BG = _ESCAPE + "41m"
    GREEN_BG = _ESCAPE + "42m"
    YELLOW_BG = _ESCAPE + "43m"
    BLUE_BG = _ESCAPE + "44m"
    MAGENTA_BG = _ESCAPE + "45m"
    CYAN_BG = _ESCAPE + "46m"
    WHITE_BG = _ESCAPE + "47m"

    BRIGHT_BLACK_FG = _ESCAPE + "90m"
    BRIGHT_RED_FG = _ESCAPE + "91m"
    BRIGHT_GREEN_FG = _ESCAPE + "92m"
    BRIGHT_YELLOW_FG = _ESCAPE + "93m"
    BRIGHT_BLUE_FG = _ESCAPE + "94m"
    BRIGHT_MAGENTA_FG = _ESCAPE + "95m"
    BRIGHT_CYAN_FG = _ESCAPE + "96m"
    BRIGHT_WHITE_FG = _ESCAPE + "97m"

    BRIGHT_BLACK_BG = _ESCAPE + "100m"
    BRIGHT_RED_BG = _ESCAPE + "101m"
    BRIGHT_GREEN_BG = _ESCAPE + "102m"
    BRIGHT_YELLOW_BG = _ESCAPE + "103m"
    BRIGHT_BLUE_BG = _ESCAPE + "104m"
    BRIGHT_MAGENTA_BG = _ESCAPE + "105m"
    BRIGHT_CYAN_BG = _ESCAPE + "106m"
    BRIGHT_WHITE_BG = _ESCAPE + "107m"



# Copyright cocotb contributors
# Licensed under the Revised BSD License, see LICENSE for details.
# SPDX-License-Identifier: BSD-3-Clause
import warnings

from cocotb._base_triggers import Event, Lock, NullTrigger, Trigger
from cocotb._extended_awaitables import (
    ClockCycles,
    Combine,
    First,
    SimTimeoutError,
    Waitable,
    with_timeout,
)
from cocotb._gpi_triggers import (
    Edge,
    FallingEdge,
    GPITrigger,
    NextTimeStep,
    ReadOnly,
    ReadWrite,
    RisingEdge,
    Timer,
    ValueChange,
    current_gpi_trigger,
)

__all__ = (
    "ClockCycles",
    "Combine",
    "Edge",
    "Event",
    "FallingEdge",
    "First",
    "GPITrigger",
    "Lock",
    "NextTimeStep",
    "NullTrigger",
    "ReadOnly",
    "ReadWrite",
    "RisingEdge",
    "SimTimeoutError",
    "Timer",
    "Trigger",
    "ValueChange",
    "Waitable",
    "current_gpi_trigger",
    "with_timeout",
)

# Set __module__ on re-exports
for name in __all__:
    obj = globals()[name]
    obj.__module__ = __name__


def __getattr__(name: str) -> object:
    if name == "Join":
        warnings.warn(
            "Join has been moved to `cocotb.task`.",
            DeprecationWarning,
            stacklevel=2,
        )
        from cocotb.task import Join  # noqa: PLC0415

        return Join
    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")



# Copyright cocotb contributors
# Copyright (c) 2013 Potential Ventures Ltd
# Copyright (c) 2013 SolarFlare Communications Inc
# Licensed under the Revised BSD License, see LICENSE for details.
# SPDX-License-Identifier: BSD-3-Clause

"""A collection of triggers which a testbench can :keyword:`await`."""

import logging
import warnings
from typing import (
    AsyncContextManager,
    Callable,
    Generator,
    List,
    Optional,
    Union,
)

from cocotb._deprecation import deprecated
from cocotb._py_compat import Self, cached_property
from cocotb._utils import pointer_str


class Trigger:
    """A future event that a Task can wait upon."""

    def __init__(self) -> None:
        self._primed = False

    @cached_property
    def _log(self) -> logging.Logger:
        return logging.getLogger(f"cocotb.{type(self).__qualname__}.0x{id(self):x}")

    def _prime(self, callback: Callable[["Self"], None]) -> None:
        """Set a callback to be invoked when the trigger fires.

        The callback will be invoked with a single argument, `self`.

        Sub-classes must override this, but should end by calling the base class
        method.

        .. warning::
            Do not call this directly within a :term:`task`. It is intended to be used
            only by the scheduler.
        """
        # Set _primed so the trigger can test if it's already been primed and behave appropriately.
        self._primed = True

    def _unprime(self) -> None:
        """Remove the callback, and perform cleanup if necessary.

        After being un-primed, a Trigger may be re-primed again in the future.
        Calling `_unprime` multiple times is allowed, subsequent calls should be
        a no-op.

        Sub-classes may override this, but should end by calling the base class
        method.

        .. warning::
            Do not call this directly within a :term:`task`. It is intended to be used
            only by the scheduler.
        """
        self._cleanup()

    def _cleanup(self) -> None:
        # Clear _primed so this Trigger can be re-primed.
        self._primed = False

    def __await__(self) -> Generator["Self", None, "Self"]:
        yield self
        return self


class _Event(Trigger):
    """Unique instance used by the Event object.

    One created for each attempt to wait on the event so that the scheduler
    can maintain a unique mapping of triggers to tasks.
    """

    _callback: Callable[["_Event"], None]

    def __init__(self, parent: "Event") -> None:
        super().__init__()
        self._parent = parent

    def _prime(self, callback: Callable[["_Event"], None]) -> None:
        if self._primed:
            return
        if self._parent.is_set():
            # If the event is already set, we need to call the callback
            # immediately, so we don't need to wait for the scheduler.
            callback(self)
            return
        self._callback = callback
        return super()._prime(callback)

    def _unprime(self) -> None:
        if not self._primed:
            return
        return super()._unprime()

    def _set(self) -> None:
        if self._primed:
            self._callback(self)

    def __repr__(self) -> str:
        return f"<{self._parent!r}.wait() at {pointer_str(self)}>"


class Event:
    r"""A way to signal an event across :class:`~cocotb.task.Task`\ s.

    :keyword:`await`\ ing the result of :meth:`wait()` will block the :keyword:`await`\ ing :class:`~cocotb.task.Task`
    until :meth:`set` is called.

    Args:
        name: Name for the Event.

    Usage:
        .. code-block:: python

            e = Event()


            async def task1():
                await e.wait()
                print("resuming!")


            cocotb.start_soon(task1())
            # do stuff
            e.set()
            await NullTrigger()  # allows task1 to execute
            # resuming!

    .. versionremoved:: 2.0

        Removed the undocumented *data* attribute and argument to :meth:`set`,
        and the *name* attribute and argument to the constructor.
    """

    def __init__(self, name: Optional[str] = None) -> None:
        self._event: _Event = _Event(self)
        self._name: Union[str, None] = None
        if name is not None:
            warnings.warn(
                "The 'name' argument will be removed in a future release.",
                DeprecationWarning,
                stacklevel=2,
            )
            self.name = name
        self._fired: bool = False
        self._data: object = None

    @property
    @deprecated("The 'name' field will be removed in a future release.")
    def name(self) -> Union[str, None]:
        """Name of the Event.

        .. deprecated:: 2.0
            The *name* field will be removed in a future release.
        """
        return self._name

    @name.setter
    @deprecated("The 'name' field will be removed in a future release.")
    def name(self, new_name: Union[str, None]) -> None:
        self._name = new_name

    @property
    @deprecated("The data field will be removed in a future release.")
    def data(self) -> object:
        """The data associated with the Event.

        .. deprecated:: 2.0
            The data field will be removed in a future release.
            Use a separate variable to store the data instead.
        """
        return self._data

    @data.setter
    @deprecated("The data field will be removed in a future release.")
    def data(self, new_data: object) -> None:
        self._data = new_data

    def set(self, data: Optional[object] = None) -> None:
        """Set the Event and unblock all Tasks blocked on this Event."""
        self._fired = True
        if data is not None:
            warnings.warn(
                "The data field will be removed in a future release.",
                DeprecationWarning,
                stacklevel=2,
            )
        self._data = data
        self._event._set()

    def wait(self) -> Trigger:
        """Block the current Task until the Event is set.

        If the event has already been set, the trigger will fire immediately.

        To set the Event call :meth:`set`.
        To reset the Event (and enable the use of :meth:`wait` again),
        call :meth:`clear`.
        """
        return self._event

    def clear(self) -> None:
        """Clear this event that has been set.

        Subsequent calls to :meth:`~cocotb.triggers.Event.wait` will block until
        :meth:`~cocotb.triggers.Event.set` is called again.
        """
        self._fired = False

    def is_set(self) -> bool:
        """Return ``True`` if event has been set."""
        return self._fired

    def __repr__(self) -> str:
        if self._name is None:
            fmt = "<{0} at {2}>"
        else:
            fmt = "<{0} for {1} at {2}>"
        return fmt.format(type(self).__qualname__, self._name, pointer_str(self))


class _InternalEvent(Trigger):
    """Event used internally for triggers that need cross-:class:`~cocotb.task.Task` synchronization.

    This Event can only be waited on once, by a single :class:`~cocotb.task.Task`.

    Provides transparent :func`repr` pass-through to the :class:`Trigger` using this event,
    providing a better debugging experience.
    """

    def __init__(self, parent: object) -> None:
        super().__init__()
        self._parent = parent
        self._callback: Optional[Callable[[_InternalEvent], None]] = None
        self.fired: bool = False

    def _prime(self, callback: Callable[["_InternalEvent"], None]) -> None:
        if self._primed:
            raise RuntimeError("This Trigger may only be awaited once")
        self._callback = callback
        super()._prime(callback)
        if self.fired:
            self._callback(self)

    def _cleanup(self) -> None:
        # Don't clear _primed so a second call to _prime() fails.
        pass

    def set(self) -> None:
        """Wake up coroutine blocked on this event."""
        self.fired = True

        if self._callback is not None:
            self._callback(self)

    def is_set(self) -> bool:
        """Return true if event has been set."""
        return self.fired

    def __await__(
        self,
    ) -> Generator["Self", None, "Self"]:
        if self._primed:
            raise RuntimeError("Only one Task may await this Trigger")
        yield self
        return self

    def __repr__(self) -> str:
        return repr(self._parent)


class _Lock(Trigger):
    """Unique instance used by the Lock object.

    One created for each attempt to acquire the Lock so that the scheduler
    can maintain a unique mapping of triggers to tasks.
    """

    def __init__(self, parent: "Lock") -> None:
        super().__init__()
        self._parent = parent

    def _prime(self, callback: Callable[["Self"], None]) -> None:
        if self._primed:
            raise RuntimeError(
                "Lock.acquire() result can only be used by one task at a time"
            )
        self._callback = callback
        self._parent._prime_lock(self)
        return super()._prime(callback)

    def _unprime(self) -> None:
        if not self._primed:
            return
        self._parent._unprime_lock(self)
        return super()._unprime()

    def __repr__(self) -> str:
        return f"<{self._parent!r}.acquire() at {pointer_str(self)}>"


class Lock(AsyncContextManager[None]):
    """A mutual exclusion lock.

    Guarantees fair scheduling.
    Lock acquisition is given in order of attempted lock acquisition.

    Usage:
        By directly calling :meth:`acquire` and :meth:`release`.

        .. code-block:: python

            lock = Lock()
            ...
            await lock.acquire()
            try:
                # do some stuff
                ...
            finally:
                lock.release()

        Or...

        .. code-block:: python

            async with Lock():
                # do some stuff
                ...

    .. versionchanged:: 1.4

        The lock can be used as an asynchronous context manager in an
        :keyword:`async with` statement
    """

    def __init__(self, name: Optional[str] = None) -> None:
        self._pending_primed: List[_Lock] = []
        self._name: Union[str, None] = None
        if name is not None:
            warnings.warn(
                "The 'name' argument will be removed in a future release.",
                DeprecationWarning,
                stacklevel=2,
            )
            self._name = name
        self._locked: bool = False

    @property
    @deprecated("The 'name' field will be removed in a future release.")
    def name(self) -> Union[str, None]:
        """Name of the Lock.

        .. deprecated:: 2.0
            The *name* field will be removed in a future release.
        """
        return self._name

    @name.setter
    @deprecated("The 'name' field will be removed in a future release.")
    def name(self, new_name: Union[str, None]) -> None:
        self._name = new_name

    def locked(self) -> bool:
        """Return ``True`` if the lock has been acquired.

        .. versionchanged:: 2.0
            This is now a method to match :meth:`asyncio.Lock.locked`, rather than an attribute.
        """
        return self._locked

    def _acquire_and_fire(self, lock: _Lock) -> None:
        self._locked = True
        lock._callback(lock)

    def _prime_lock(self, lock: _Lock) -> None:
        if not self._locked:
            self._acquire_and_fire(lock)
        else:
            self._pending_primed.append(lock)

    def _unprime_lock(self, lock: _Lock) -> None:
        if lock in self._pending_primed:
            self._pending_primed.remove(lock)

    def acquire(self) -> Trigger:
        """Produce a trigger which fires when the lock is acquired."""
        trig = _Lock(self)
        return trig

    def release(self) -> None:
        """Release the lock."""
        if not self._locked:
            raise RuntimeError(f"Attempt to release an unacquired Lock {self!s}")

        self._locked = False

        # nobody waiting for this lock
        if not self._pending_primed:
            return

        lock = self._pending_primed.pop(0)
        self._acquire_and_fire(lock)

    def __repr__(self) -> str:
        if self._name is None:
            fmt = "<{0} [{2} waiting] at {3}>"
        else:
            fmt = "<{0} for {1} [{2} waiting] at {3}>"
        return fmt.format(
            type(self).__qualname__,
            self._name,
            len(self._pending_primed),
            pointer_str(self),
        )

    async def __aenter__(self) -> None:
        await self.acquire()

    async def __aexit__(self, *args: object) -> None:
        self.release()


class NullTrigger(Trigger):
    """Trigger that fires immediately.

    Mostly useful when building or using higher-order functions which need to take or return Triggers.

    The scheduling order of the Task awaiting this Trigger with respect to any other Task is not deterministic
    and should generally not be relied upon.
    Instead of using this Trigger to push the Task until "after" another Task has run,
    use other synchronization techniques, such as using an :class:`.Event`.

    **Do not** do this:

    .. code-block:: python
        :class: removed

        transaction_data = None


        def monitor(dut):
            while dut.valid.value != 1 and dut.ready.value != 1:
                await RisingEdge(dut.clk)
            transaction_data = dut.data.value


        def use_transaction(dut):
            while True:
                await RisingEdge(dut.clk)
                # We need the NullTrigger here because both Tasks react to RisingEdge,
                # but there's no guarantee about which Task is run first,
                # so we need to force this one to run "later" using NullTrigger.
                await NullTrigger()
                if transaction_data is not None:
                    process(transaction_data)


        use_task = cocotb.start_soon(use_transaction(cocotb.top))
        monitor_task = cocotb.start_soon(monitor(cocotb.top))

    Instead use an :class:`!Event` to explicitly synchronize the two Tasks, like so:

    .. code-block:: python
        :class: new

        transaction_data = None
        transaction_event = Event()


        def monitor(dut):
            while dut.valid.value != 1 and dut.ready.value != 1:
                await RisingEdge(dut.clk)
            transaction_data = dut.data.value
            transaction_event.set()


        def use_transaction(dut):
            # Now we don't need the NullTrigger.
            # This Task will wake up *strictly* after `monitor_task` sets the transaction.
            await transaction_event.wait()
            process(transaction_data)


        use_task = cocotb.start_soon(use_transaction(cocotb.top))
        monitor_task = cocotb.start_soon(monitor(cocotb.top))

    .. versionremoved:: 2.0
        The *outcome* parameter was removed. There is no alternative.
    """

    def __init__(self, name: Optional[str] = None) -> None:
        super().__init__()
        self.name = name

    def _prime(self, callback: Callable[["Self"], None]) -> None:
        if self._primed:
            return
        callback(self)
        return super()._prime(callback)

    def __repr__(self) -> str:
        if self.name is None:
            fmt = "<{0} at {2}>"
        else:
            fmt = "<{0} for {1} at {2}>"
        return fmt.format(type(self).__qualname__, self.name, pointer_str(self))



"""
Inspired by https://github.com/python-trio/outcome

An outcome is similar to the built-in :any:`concurrent.futures.Future`
or :any:`asyncio.Future`, but without being tied to a particular task model.
"""

import abc
from typing import Callable, Generic, TypeVar

from cocotb._py_compat import ParamSpec
from cocotb._utils import remove_traceback_frames

T = TypeVar("T")
P = ParamSpec("P")


def capture(
    fn: "Callable[P, T]", *args: "P.args", **kwargs: "P.kwargs"
) -> "Outcome[T]":
    """Obtain an `Outcome` representing the result of a function call."""
    try:
        return Value(fn(*args, **kwargs))
    except BaseException as e:
        e = remove_traceback_frames(e, ["capture"])
        return Error(e)


class Outcome(Generic[T]):
    @abc.abstractmethod
    def get(self) -> T:
        """Get the value of this outcome, or throw its exception."""


class Value(Outcome[T]):
    def __init__(self, value: T):
        self.value = value

    def get(self) -> T:
        return self.value

    def __repr__(self) -> str:
        return f"Value({self.value!r})"


class Error(Outcome[T]):
    def __init__(self, error: BaseException) -> None:
        self.error = error

    def get(self) -> T:
        raise self.error

    def __repr__(self) -> str:
        return f"Error({self.error!r})"



# Copyright cocotb contributors
# Licensed under the Revised BSD License, see LICENSE for details.
# SPDX-License-Identifier: BSD-3-Clause

"""Build HDL and run cocotb tests."""

# TODO: maybe do globbing and expanduser/expandvars in --include, --vhdl-sources, --verilog-sources
# TODO: create a short README and a .gitignore (content: "*") in both build_dir and test_dir? (Some other tools do this.)
# TODO: support timescale on all simulators
# TODO: support custom dependencies

import logging
import multiprocessing
import os
import re
import shlex
import shutil
import subprocess
import sys
import tempfile
import warnings
from abc import ABC, abstractmethod
from contextlib import suppress
from itertools import chain
from pathlib import Path
from typing import (
    Dict,
    Generic,
    Iterable,
    List,
    Mapping,
    Optional,
    Sequence,
    TextIO,
    Tuple,
    Type,
    TypeVar,
    Union,
)

import find_libpython

import cocotb_tools.config
from cocotb_tools.check_results import get_results
from cocotb_tools.sim_versions import NvcVersion

if sys.version_info >= (3, 10):
    from typing import TypeAlias

PathLike: "TypeAlias" = Union["os.PathLike[str]", str]
"A path that can be passed to :class:`pathlib.Path` or :func:`open`"

_Command: "TypeAlias" = List[str]

_magic_re = re.compile(r"([\\{}])")
_space_re = re.compile(r"([\s])", re.ASCII)


MAX_PARALLEL_BUILD_JOBS: int = 4
"""The maximum number of parallel build threads in calls to :meth:`.Runner.build`.

If the number of CPU cores is less than this value, it uses the CPU core count.
Set this variable to globally change the number of parallel build jobs.
"""


def _get_max_parallel_build_jobs() -> int:
    return min(MAX_PARALLEL_BUILD_JOBS, multiprocessing.cpu_count())


def _as_tcl_value(value: str) -> str:
    # add '\' before special characters and spaces
    value = _magic_re.sub(r"\\\1", value)
    value = value.replace("\n", r"\n")
    value = _space_re.sub(r"\\\1", value)
    if value[:1] == '"':
        value = "\\" + value

    return value


_sv_escapes = {
    "\n": "\\n",
    "\t": "\\t",
    "\\": "\\\\",
    '"': '\\"',
    "\v": "\\v",
    "\f": "\\f",
    "\xff": "\\xFF",
}
for i in range(32):
    if chr(i) not in _sv_escapes:
        _sv_escapes[chr(i)] = f"\\x{i:02x}"

_sv_escape_translate_table = str.maketrans(_sv_escapes)


def _as_sv_literal(value: object) -> str:
    if isinstance(value, (int, float)):
        return str(value)
    elif isinstance(value, str):
        return '"' + value.translate(_sv_escape_translate_table) + '"'
    else:
        raise TypeError("Can't serialize this type as an SV literal")


def _shlex_join(split_command: Iterable[str]) -> str:
    """
    Return a shell-escaped string from *split_command*
    This is here more for compatibility purposes
    """
    return " ".join(shlex.quote(arg) for arg in split_command)


_T = TypeVar("_T")


class _Tag(Generic[_T]):
    def __init__(self, value: _T) -> None:
        self.value = value

    def __repr__(self) -> str:
        return f"{type(self).__qualname__}({self.value!r})"


class _ValueAndTag(Generic[_T]):
    def __init__(self, value: _T, tag: Type[_Tag]) -> None:
        self.value = value
        self.tag = tag


class _ValueAndOptionalTag(Generic[_T]):
    def __init__(self, value: _T, tag: Union[Type[_Tag], None]) -> None:
        self.value = value
        self.tag = tag


class VHDL(_Tag):
    """Tags source files and build arguments to :meth:`Runner.build() <cocotb_tools.runner.Runner.build>` as VHDL-specific."""


class Verilog(_Tag):
    """Tags source files and build arguments to :meth:`Runner.build() <cocotb_tools.runner.Runner.build>` as Verilog-specific."""


class VerilatorControlFile(_Tag):
    """Tags source files to :meth:`Runner.build() <cocotb_tools.runner.Runner.build>` as Verilator control files."""


_verilog_extensions = (".v", ".sv", ".vh", ".svh")
_vhdl_extensions = (".vhd", ".vhdl")


def _determine_file_type(
    filename: PathLike,
) -> Union[Type[Verilog], Type[VHDL], Type[VerilatorControlFile]]:
    ext = Path(filename).suffix
    if ext in _verilog_extensions:
        return Verilog
    elif ext in _vhdl_extensions:
        return VHDL
    elif ext == "vlt":
        return VerilatorControlFile
    else:
        raise ValueError(
            f"Can't determine source file type of {filename}. Use the `VHDL`, `Verilog`, or `VerilatorControlFile` tags."
        )


class Runner(ABC):
    supported_gpi_interfaces: Dict[str, List[str]] = {}

    def __init__(self) -> None:
        self._simulator_in_path()

        self.env: Dict[str, str] = {}

        # for running test() independently of build()
        self.build_dir: Path = get_abs_path("sim_build")
        self.parameters: Mapping[str, object] = {}

        self.log = logging.getLogger(type(self).__qualname__)
        self.log.setLevel(logging.INFO)

    @abstractmethod
    def _simulator_in_path(self) -> None:
        """Raise exception if the simulator executable does not exist in :envvar:`PATH`.

        Raises:
            SystemExit: Simulator executable does not exist in :envvar:`PATH`.
        """

    def _check_hdl_toplevel_lang(self, hdl_toplevel_lang: Optional[str]) -> str:
        """Return *hdl_toplevel_lang* if supported by simulator, raise exception otherwise.

        Returns:
            *hdl_toplevel_lang* if supported by the simulator.

        Raises:
            ValueError: *hdl_toplevel_lang* is not supported by the simulator.
        """
        if hdl_toplevel_lang is None:
            if self._vhdl_sources and not self._verilog_sources and not self._sources:
                lang = "vhdl"
            elif self._verilog_sources and not self._vhdl_sources and not self._sources:
                lang = "verilog"
            elif self._sources and not self._vhdl_sources and not self._verilog_sources:
                top_source = self._sources[-1]
                if top_source.tag is VHDL:
                    lang = "vhdl"
                elif top_source.tag is Verilog:
                    lang = "verilog"
                else:
                    raise ValueError(
                        "First argument to *sources* must be a VHDL or Verilog file. "
                        f"Got a {top_source.tag.__qualname__} file: {top_source.value}"
                    )
            else:
                raise ValueError(
                    f"{type(self).__qualname__}: Must specify a hdl_toplevel_lang in a mixed-language design"
                )
        else:
            lang = hdl_toplevel_lang

        if lang in self.supported_gpi_interfaces:
            return lang
        else:
            raise ValueError(
                f"{type(self).__qualname__}: hdl_toplevel_lang {hdl_toplevel_lang!r} is not "
                f"in supported list: {', '.join(self.supported_gpi_interfaces)}"
            )

    def _set_env(self) -> None:
        """Set environment variables for sub-processes."""

        for e in os.environ:
            self.env[e] = os.environ[e]

        if "LIBPYTHON_LOC" not in self.env:
            libpython_path = find_libpython.find_libpython()
            if not libpython_path:
                raise ValueError(
                    "Unable to find libpython, please make sure the appropriate libpython is installed"
                )
            self.env["LIBPYTHON_LOC"] = libpython_path

        self.env["PATH"] += os.pathsep + str(cocotb_tools.config.libs_dir)
        self.env["PYTHONPATH"] = os.pathsep.join(sys.path)
        self.env["PYGPI_PYTHON_BIN"] = sys.executable
        self.env["COCOTB_TOPLEVEL"] = self.sim_hdl_toplevel
        self.env["COCOTB_TEST_MODULES"] = self.test_module
        self.env["TOPLEVEL_LANG"] = self.hdl_toplevel_lang

    @abstractmethod
    def _build_command(self) -> Sequence[_Command]:
        """Return command to build the HDL sources."""

    @abstractmethod
    def _test_command(self) -> Sequence[_Command]:
        """Return command to run a test."""

    def _use_external_viewer(self) -> bool:
        """Return if an external viewer should be called after simulation when ``gui=True``."""
        return False

    def _waves_file(self) -> Optional[str]:
        """Return file name of the generated waveform file for use with external viewer."""
        return None

    def build(
        self,
        hdl_library: str = "top",
        verilog_sources: Sequence[Union[PathLike, Verilog]] = [],
        vhdl_sources: Sequence[Union[PathLike, VHDL]] = [],
        sources: Sequence[Union[PathLike, VHDL, Verilog, VerilatorControlFile]] = [],
        includes: Sequence[PathLike] = [],
        defines: Mapping[str, object] = {},
        parameters: Mapping[str, object] = {},
        build_args: Sequence[Union[str, VHDL, Verilog]] = [],
        hdl_toplevel: Optional[str] = None,
        always: bool = False,
        build_dir: PathLike = "sim_build",
        clean: bool = False,
        verbose: bool = False,
        timescale: Optional[Tuple[str, str]] = None,
        waves: bool = False,
        log_file: Optional[PathLike] = None,
    ) -> None:
        """Build the HDL sources.

        With mixed language simulators, *sources* will be built,
        followed by *vhdl_sources*, then *verilog_sources*.
        With simulators that only support either VHDL or Verilog, *sources* will be built,
        followed by *vhdl_sources* and *verilog_sources*, respectively.

        If your source files use an atypical file extension,
        use :class:`VHDL`, :class:`Verilog`, or :class:`VerilatorControlFile`
        to tag the path as a VHDL, Verilog, or Verilator control file source file, respectively.
        If the filepaths aren't tagged, the extension is used to determine if they are VHDL or Verilog files.

        +----------+------------------------------------+
        | Language | File Extensions                    |
        +==========+====================================+
        | VHDL     | ``.vhd``, ``.vhdl``                |
        +----------+------------------------------------+
        | Verilog  | ``.v``, ``.sv``, ``.vh``, ``.svh`` |
        +----------+------------------------------------+


        .. code-block:: python

            runner.build(
                sources=[
                    VHDL("/my/file.is_actually_vhdl"),
                    Verilog("/other/file.verilog"),
                ],
            )

        The same tagging works for *build_args*.
        Tagged *build_args* only supply that option to the compiler when building the source file for the tagged language.
        Non-tagged *build_args* are supplied when compiling any language.

        Args:
            hdl_library: The library name to compile into.
            verilog_sources: Verilog source files to build.
            vhdl_sources: VHDL source files to build.
            sources: Language-agnostic list of source files to build.
            includes: Verilog include directories.
            defines: Defines to set.
            parameters: Verilog parameters or VHDL generics.
            build_args: Extra build arguments for the simulator.
            hdl_toplevel: The name of the HDL toplevel module.
            always: Always run the build step.
            build_dir: Directory to run the build step in.
            clean: Delete *build_dir* before building.
            verbose: Enable verbose messages.
            timescale: Tuple containing time unit and time precision for simulation.
            waves: Record signal traces. Overridden by the :envvar:`WAVES` environment variable.
            log_file: File to write the build log to.

        .. deprecated:: 2.0

            Uses of the *verilog_sources* and *vhdl_sources* parameters should be replaced with the language-agnostic *sources* argument.
        """

        self.clean: bool = clean
        self.build_dir = get_abs_path(build_dir)
        if self.clean:
            self.rm_build_folder(self.build_dir)
        self.build_dir.mkdir(parents=True, exist_ok=True)

        # note: to avoid mutating argument defaults, we ensure that no value
        # is written without a copy. This is much more concise and leads to
        # a better docstring than using `None` as a default in the parameters
        # list.
        self.hdl_library: str = hdl_library
        if verilog_sources:
            warnings.warn(
                "Simulator.build *verilog_sources* parameter is deprecated. Use the language-agnostic *sources* parameter instead.",
                DeprecationWarning,
                stacklevel=2,
            )
        self._set_verilog_sources(verilog_sources)
        if vhdl_sources:
            warnings.warn(
                "Simulator.build *vhdl_sources* parameter is deprecated. Use the language-agnostic *sources* parameter instead.",
                DeprecationWarning,
                stacklevel=2,
            )
        self._set_vhdl_sources(vhdl_sources)
        self._set_sources(sources)
        self.includes: List[Path] = [
            get_abs_path(include_dir) for include_dir in includes
        ]
        self.defines = dict(defines)
        self.parameters = dict(parameters)
        self._set_build_args(build_args)
        self.always: bool = always
        self.hdl_toplevel: Optional[str] = hdl_toplevel
        self.verbose: bool = verbose
        self.timescale: Optional[Tuple[str, str]] = timescale
        self.log_file: Optional[PathLike] = log_file

        self.waves = bool(os.getenv("WAVES", waves))

        self.env.update(os.environ)

        cmds: Sequence[_Command] = self._build_command()
        self._execute(cmds, cwd=self.build_dir)

    def test(
        self,
        test_module: Union[str, Sequence[str]],
        hdl_toplevel: str,
        hdl_toplevel_library: str = "top",
        hdl_toplevel_lang: Optional[str] = None,
        gpi_interfaces: Optional[List[str]] = None,
        testcase: Optional[Union[str, Sequence[str]]] = None,
        seed: Optional[Union[str, int]] = None,
        elab_args: Sequence[str] = [],
        test_args: Sequence[str] = [],
        plusargs: Sequence[str] = [],
        extra_env: Mapping[str, str] = {},
        waves: bool = False,
        gui: bool = False,
        parameters: Optional[Mapping[str, object]] = None,
        build_dir: Optional[PathLike] = None,
        test_dir: Optional[PathLike] = None,
        results_xml: Optional[str] = None,
        pre_cmd: Optional[List[str]] = None,
        verbose: bool = False,
        timescale: Optional[Tuple[str, str]] = None,
        log_file: Optional[PathLike] = None,
        test_filter: Optional[str] = None,
    ) -> Path:
        """Run the tests.

        Args:
            test_module: Name(s) of the Python module(s) containing the tests to run.
                Can be a comma-separated list.
            hdl_toplevel: Name of the HDL toplevel module.
            hdl_toplevel_library: The library name for HDL toplevel module.
            hdl_toplevel_lang: Language of the HDL toplevel module.
            gpi_interfaces: List of GPI interfaces to use, with the first one being the entry point.
            testcase: Name(s) of a specific testcase(s) to run.
                If not set, run all testcases found in *test_module*.
                Can be a comma-separated list.
            seed: A specific random seed to use.
            elab_args: A list of elaboration arguments for the simulator.
            test_args: A list of extra arguments for the simulator.
            plusargs: 'plusargs' to set for the simulator.
            extra_env: Extra environment variables to set.
            waves: Record signal traces. Overridden by the :envvar:`WAVES` environment variable.
            gui: Run with simulator GUI. Overridden by the :envvar:`GUI` environment variable.
            parameters: Verilog parameters or VHDL generics.
            build_dir: Directory the build step has been run in.
            test_dir: Directory to run the tests in.
            results_xml: Name of xUnit XML file to store test results in.
                If an absolute path is provided it will be used as-is,
                :file:`{build_dir}/results.xml` otherwise.
                This argument should not be set when run with ``pytest``.
            verbose: Enable verbose messages.
            pre_cmd: Commands to run before simulation begins.
                Typically Tcl commands for simulators that support them.
            timescale: Tuple containing time unit and time precision for simulation.
            log_file: File to write the test log to.
            test_filter: Regular expression which matches test names.
                Only matched tests are run if this argument if given.

        Returns:
            The absolute location of the results XML file which can be
            defined by the *results_xml* argument.
        """

        __tracebackhide__ = True  # Hide the traceback when using pytest

        if build_dir is not None:
            self.build_dir = get_abs_path(build_dir)

        if parameters is not None:
            self.parameters = dict(parameters)

        if test_dir is None:
            self.test_dir = self.build_dir
        else:
            self.test_dir = get_abs_path(test_dir)
        self.test_dir.mkdir(parents=True, exist_ok=True)

        if isinstance(test_module, str):
            self.test_module = test_module
        else:
            self.test_module = ",".join(test_module)

        # note: to avoid mutating argument defaults, we ensure that no value
        # is written without a copy. This is much more concise and leads to
        # a better docstring than using `None` as a default in the parameters
        # list.
        self.sim_hdl_toplevel = hdl_toplevel
        self.hdl_toplevel_library: str = hdl_toplevel_library
        self.hdl_toplevel_lang = self._check_hdl_toplevel_lang(hdl_toplevel_lang)
        if gpi_interfaces:
            self.gpi_interfaces = gpi_interfaces
        else:
            self.gpi_interfaces = []
            for gpi_if in self.supported_gpi_interfaces.values():
                self.gpi_interfaces.append(gpi_if[0])

        self.pre_cmd = pre_cmd

        self.elab_args = list(elab_args)
        self.test_args = list(test_args)
        self.plusargs = list(plusargs)
        self.env = dict(extra_env)

        if testcase is not None:
            if isinstance(testcase, str):
                self.env["COCOTB_TESTCASE"] = testcase
            else:
                self.env["COCOTB_TESTCASE"] = ",".join(testcase)

        if test_filter is not None:
            self.env["COCOTB_TEST_FILTER"] = test_filter

        if seed is not None:
            self.env["COCOTB_RANDOM_SEED"] = str(seed)

        self.log_file = log_file
        self.waves = bool(int(os.getenv("WAVES", waves)))
        self.gui = bool(int(os.getenv("GUI", gui)))
        self.timescale = timescale

        if verbose is not None:
            self.verbose = verbose

        # Pytest test name is used by the next couple sections.
        pytest_current_test = os.getenv("PYTEST_CURRENT_TEST", None)

        if pytest_current_test is not None:
            self.current_test_name = pytest_current_test.split(":")[-1].split(" ")[0]
        else:
            self.current_test_name = "test"

        results_xml_path: Union[None, Path] = (
            Path(results_xml) if results_xml is not None else None
        )

        # result.xml filename precedence:
        # 1. absolute path
        # 2. pytest test name
        # 3. relative path
        # 4. default name
        if results_xml_path is not None and results_xml_path.is_absolute():
            results_xml_file = results_xml_path
        elif pytest_current_test is not None:
            if results_xml_path is not None:
                raise NotImplementedError(
                    "Relative result_xml paths aren't supported when using pytest"
                )
            results_xml_file = self.test_dir / f"{self.current_test_name}.result.xml"
        elif results_xml_path is not None:
            results_xml_file = self.test_dir / results_xml_path
        else:
            results_xml_file = self.test_dir / "results.xml"

        with suppress(OSError):
            results_xml_file.unlink()

        # transport the settings to cocotb via environment variables
        self._set_env()
        self.env["COCOTB_RESULTS_FILE"] = str(results_xml_file)

        cmds: Sequence[_Command] = self._test_command()
        simulator_exit_code: int = 0
        try:
            self._execute(cmds, cwd=self.test_dir)
        except subprocess.CalledProcessError as e:
            # It is possible for the simulator to fail but still leave results.
            self.log.error("Simulation failed: %d", e.returncode)
            simulator_exit_code = e.returncode

        # Only when running under pytest, check the results file here,
        # potentially raising an exception with failing testcases,
        # otherwise return the results file for later analysis.
        if pytest_current_test:
            try:
                (num_tests, num_failed) = get_results(results_xml_file)
            except RuntimeError as e:
                self.log.error("%s", e.args[0])
                sys.exit(simulator_exit_code)
            else:
                if num_failed:
                    self.log.error(
                        "ERROR: Failed %d of %d tests.", num_failed, num_tests
                    )
                    sys.exit(1 if simulator_exit_code == 0 else simulator_exit_code)

        if simulator_exit_code != 0:
            sys.exit(simulator_exit_code)

        if pytest_current_test and self._use_external_viewer() and self.gui:
            viewer = os.getenv("COCOTB_WAVEFORM_VIEWER")
            if viewer is not None:
                viewer_path = shutil.which(viewer)
                if viewer_path is None:
                    raise ValueError(f"Cannot find {viewer} in the system path")
            else:
                viewer_path = shutil.which("surfer")
                if viewer_path is None:
                    viewer_path = shutil.which("gtkwave")
            if viewer_path is None:
                raise SystemError(
                    "Cannot find any viewer (surfer or gtkwave) in the system path"
                )

            subprocess.run(
                [f"{viewer_path} {self._waves_file()}"],
                cwd=self.test_dir,
                check=True,
                shell=True,
            )

        self.log.info("Results file: %s", results_xml_file)
        return results_xml_file

    @abstractmethod
    def _get_include_options(self, includes: Sequence[PathLike]) -> _Command:
        """Return simulator-specific formatted option strings with *includes* directories."""

    @abstractmethod
    def _get_define_options(self, defines: Mapping[str, object]) -> _Command:
        """Return simulator-specific formatted option strings with *defines* macros."""

    @abstractmethod
    def _get_parameter_options(self, parameters: Mapping[str, object]) -> _Command:
        """Return simulator-specific formatted option strings with *parameters*/generics."""

    def _execute(self, cmds: Sequence[_Command], cwd: PathLike) -> None:
        __tracebackhide__ = True  # Hide the traceback when using PyTest.

        if self.log_file is None:
            self._execute_cmds(cmds, cwd)
        else:
            with open(self.log_file, "w") as f:
                self._execute_cmds(cmds, cwd, f)

    def _execute_cmds(
        self, cmds: Sequence[_Command], cwd: PathLike, stdout: Optional[TextIO] = None
    ) -> None:
        __tracebackhide__ = True  # Hide the traceback when using PyTest.

        for cmd in cmds:
            self.log.info("Running command %s in directory %s", _shlex_join(cmd), cwd)

            # TODO: create a thread to handle stderr and log as error?
            # TODO: log forwarding

            stderr = None if stdout is None else subprocess.STDOUT
            subprocess.run(
                cmd, cwd=cwd, env=self.env, check=True, stdout=stdout, stderr=stderr
            )

    def rm_build_folder(self, build_dir: Path) -> None:
        if build_dir.is_dir():
            self.log.info("Removing: %s", build_dir)
            shutil.rmtree(build_dir, ignore_errors=True)

    @property
    def verilog_sources(self) -> List[Path]:
        return [source.value for source in self._verilog_sources]

    @verilog_sources.setter
    def verilog_sources(self, value: List[Path]) -> None:
        self._set_verilog_sources(value)

    def _set_verilog_sources(
        self, sources: Sequence[Union[PathLike, Verilog, VerilatorControlFile]]
    ) -> None:
        verilog_sources: List[_ValueAndTag] = []
        for source in sources:
            if isinstance(source, _Tag):
                if isinstance(source, (Verilog, VerilatorControlFile)):
                    abs_path = get_abs_path(source.value)
                    verilog_sources.append(_ValueAndTag(abs_path, type(source)))
                else:
                    raise ValueError(f"Unsupported file type: {source}")
            else:
                tag = _determine_file_type(source)
                abs_path = get_abs_path(source)
                verilog_sources.append(_ValueAndTag(abs_path, tag))
        self._verilog_sources = verilog_sources

    @property
    def vhdl_sources(self) -> List[Path]:
        return [source.value for source in self._vhdl_sources]

    @vhdl_sources.setter
    def vhdl_sources(self, value: List[Path]) -> None:
        self._set_vhdl_sources(value)

    def _set_vhdl_sources(self, sources: Sequence[Union[PathLike, VHDL]]) -> None:
        vhdl_sources: List[_ValueAndTag] = []
        for source in sources:
            if isinstance(source, _Tag):
                if isinstance(source, VHDL):
                    abs_path = get_abs_path(source.value)
                    vhdl_sources.append(_ValueAndTag(abs_path, type(source)))
                else:
                    raise ValueError(f"Unsupported file type: {source}")
            else:
                tag = _determine_file_type(source)
                abs_path = get_abs_path(source)
                vhdl_sources.append(_ValueAndTag(abs_path, tag))
        self._vhdl_sources = vhdl_sources

    @property
    def sources(self) -> List[Path]:
        return [source.value for source in self._sources]

    @sources.setter
    def sources(self, value: List[Path]) -> None:
        self._set_sources(value)

    def _set_sources(
        self, sources: Sequence[Union[PathLike, Verilog, VHDL, VerilatorControlFile]]
    ) -> None:
        sources_: List[_ValueAndTag] = []
        for source in sources:
            if isinstance(source, _Tag):
                if isinstance(source, (Verilog, VHDL, VerilatorControlFile)):
                    abs_path = get_abs_path(source.value)
                    sources_.append(_ValueAndTag(abs_path, type(source)))
                else:
                    raise ValueError(f"Unsupported file type: {source}")
            else:
                tag = _determine_file_type(source)
                abs_path = get_abs_path(source)
                sources_.append(_ValueAndTag(abs_path, tag))
        self._sources = sources_

    @property
    def build_args(self) -> List[str]:
        return [arg.value for arg in self._build_args]

    @build_args.setter
    def build_args(self, value: List[str]) -> None:
        self._set_build_args(value)

    def _set_build_args(self, build_args: Sequence[Union[str, Verilog, VHDL]]) -> None:
        build_args_: List[_ValueAndOptionalTag] = []
        for build_arg in build_args:
            if isinstance(build_arg, _Tag):
                if isinstance(build_arg, (Verilog, VHDL)):
                    build_args_.append(
                        _ValueAndOptionalTag(build_arg.value, type(build_arg))
                    )
                else:
                    raise ValueError(f"Unsupported tag type: {build_arg}")
            else:
                build_args_.append(_ValueAndOptionalTag(build_arg, None))
        self._build_args = build_args_


def outdated(output: Path, dependencies: Iterable[Path]) -> bool:
    """Return ``True`` if any source files in *dependencies* are newer than the *output* directory.

    Returns:
        ``True`` if any source files are newer, ``False`` otherwise.
    """

    if not output.is_file():
        return True

    output_mtime = output.stat().st_mtime

    dep_mtime = 0.0
    for dependency in dependencies:
        mtime = dependency.stat().st_mtime
        dep_mtime = max(mtime, dep_mtime)

    return dep_mtime > output_mtime


def get_abs_path(path: PathLike) -> Path:
    """Return *path* in absolute form."""

    path = Path(path)
    if path.is_absolute():
        return path.resolve()
    else:
        return Path(Path.cwd() / path).resolve()


class Icarus(Runner):
    """Implementation of :class:`Runner` for Icarus Verilog.

    .. admonition:: Simulator-specific Usage

       * ``hdl_toplevel`` argument to :meth:`.build` is *required*.
       * ``waves=True`` *must* be given to :meth:`.build` if either ``waves`` or ``gui`` are to be used during :meth:`.test`.
       * ``timescale`` argument to :meth:`.build` must be given to support dumping the command file.
       * Does not support the ``pre_cmd`` argument to :meth:`.test`.
    """

    supported_gpi_interfaces = {"verilog": ["vpi"]}

    def _simulator_in_path(self) -> None:
        if shutil.which("iverilog") is None:
            raise SystemExit("ERROR: iverilog executable not found!")

    def _get_include_options(self, includes: Sequence[PathLike]) -> _Command:
        return [f"-I{include}" for include in includes]

    def _get_define_options(self, defines: Mapping[str, object]) -> _Command:
        return [f"-D{name}={_as_sv_literal(value)}" for name, value in defines.items()]

    def _get_parameter_options(self, parameters: Mapping[str, object]) -> _Command:
        return [
            f"-P{self.hdl_toplevel}.{name}={value}"
            for name, value in parameters.items()
        ]

    def _use_external_viewer(self) -> bool:
        return True

    def _waves_file(self) -> Optional[str]:
        return f"{self.hdl_toplevel}.fst"

    def _create_cmd_file(self) -> None:
        assert self.timescale is not None
        with open(self.cmds_file, "w") as f:
            f.write("+timescale+{}/{}\n".format(*self.timescale))

    def _create_iverilog_dump_file(self) -> None:
        dumpfile_path = _as_sv_literal(str(self.build_dir / f"{self.hdl_toplevel}.fst"))
        with open(self.iverilog_dump_file, "w") as f:
            f.write("module cocotb_iverilog_dump();\n")
            f.write("initial begin\n")
            f.write("    string dumpfile_path;")
            f.write(
                '    if ($value$plusargs("dumpfile_path=%s", dumpfile_path)) begin\n'
            )
            f.write("        $dumpfile(dumpfile_path);\n")
            f.write("    end else begin\n")
            f.write(f"        $dumpfile({dumpfile_path});\n")
            f.write("    end\n")
            f.write(f"    $dumpvars(0, {self.hdl_toplevel});\n")
            f.write("end\n")
            f.write("endmodule\n")

    @property
    def sim_file(self) -> Path:
        return self.build_dir / "sim.vvp"

    @property
    def iverilog_dump_file(self) -> Path:
        return self.build_dir / "cocotb_iverilog_dump.v"

    @property
    def cmds_file(self) -> Path:
        return self.build_dir / "cmds.f"

    def _build_command(self) -> List[_Command]:
        if self.hdl_toplevel is None:
            raise ValueError("hdl_toplevel argument is required for all Icarus builds")

        sources = self._sources + self._verilog_sources

        for source in sources:
            if source.tag is not Verilog:
                raise ValueError(
                    f"{type(self).__qualname__} only supports Verilog. {str(source.value)!r} cannot be compiled."
                )

        for arg in self._build_args:
            if arg.tag not in (Verilog, None):
                raise ValueError(
                    f"{type(self).__qualname__} only supports Verilog. build_args {arg.value!r} cannot be applied."
                )

        build_args = [arg.value for arg in self._build_args]
        if self.waves:
            self._create_iverilog_dump_file()
            build_args += ["-s", "cocotb_iverilog_dump"]

        if self.timescale is not None:
            self._create_cmd_file()
            build_args += ["-f", str(self.cmds_file)]

        cmds: list[_Command] = []
        if outdated(self.sim_file, (source.value for source in sources)) or self.always:
            cmds = [
                [
                    "iverilog",
                    "-o",
                    str(self.sim_file),
                    "-s",
                    self.hdl_toplevel,
                    "-g2012",
                ]
                + self._get_define_options(self.defines)
                + self._get_include_options(self.includes)
                + self._get_parameter_options(self.parameters)
                + [arg for arg in build_args if type(arg) in (str, Verilog)]
                + [str(source_file.value) for source_file in sources]
                + [
                    str(source_file)
                    for source_file in [self.iverilog_dump_file]
                    if self.waves
                ]
            ]

        else:
            self.log.warning("Skipping compilation of %s", self.sim_file)

        return cmds

    def _test_command(self) -> List[_Command]:
        if self.pre_cmd is not None:
            raise RuntimeError("pre_cmd is not implemented for Icarus Verilog.")

        plusargs = self.plusargs
        if self.waves or self.gui:
            plusargs += ["-fst"]
        else:
            # Disable waveform output
            plusargs += ["-none"]

        return [
            [
                "vvp",
                "-M",
                str(cocotb_tools.config.libs_dir),
                "-m",
                cocotb_tools.config.lib_name("vpi", "icarus"),
                *self.test_args,
                str(self.sim_file),
                *plusargs,
            ]
        ]


class Questa(Runner):
    """Implementation of :class:`Runner` for Siemens Questa.

    .. admonition:: Simulator-specific Usage

       * Does not support the ``timescale`` argument to :meth:`.build` or :meth:`.test`.
    """

    supported_gpi_interfaces = {"verilog": ["vpi"], "vhdl": ["fli", "vhpi"]}

    def _simulator_in_path(self) -> None:
        if shutil.which("vsim") is None:
            raise SystemExit("ERROR: vsim executable not found!")

    def _get_include_options(self, includes: Sequence[PathLike]) -> _Command:
        return [f"+incdir+{_as_tcl_value(str(include))}" for include in includes]

    def _get_define_options(self, defines: Mapping[str, object]) -> _Command:
        return [
            f"+define+{name}={_as_sv_literal(value)}" for name, value in defines.items()
        ]

    def _get_parameter_options(self, parameters: Mapping[str, object]) -> _Command:
        return [f"-g{name}={value}" for name, value in parameters.items()]

    def _build_command(self) -> List[_Command]:
        cmds = []

        cmds.append(["vlib", _as_tcl_value(self.hdl_library)])

        vhdl_args = [
            _as_tcl_value(arg.value)
            for arg in self._build_args
            if arg.tag in (VHDL, None)
        ]
        verilog_args = [
            _as_tcl_value(arg.value)
            for arg in self._build_args
            if arg.tag in (Verilog, None)
        ]
        hdl_library = _as_tcl_value(self.hdl_library)
        defines = self._get_define_options(self.defines)
        includes = self._get_include_options(self.includes)

        for source in chain(self._sources, self._vhdl_sources, self._verilog_sources):
            if source.tag is VHDL:
                cmds.append(
                    [
                        "vcom",
                        "-work",
                        hdl_library,
                        *vhdl_args,
                        _as_tcl_value(str(source.value)),
                    ]
                )
            elif source.tag is Verilog:
                cmds.append(
                    [
                        "vlog",
                        *([] if self.always else ["-incr"]),
                        "-work",
                        hdl_library,
                        "-sv",
                        *defines,
                        *includes,
                        *verilog_args,
                        _as_tcl_value(str(source.value)),
                    ]
                )
            else:
                raise ValueError(f"Unsupported file type: {source.value}")

        return cmds

    def _test_command(self) -> List[_Command]:
        cmds = []

        if self.pre_cmd is not None:
            pre_cmd = ["-do", *self.pre_cmd]
        else:
            pre_cmd = []

        do_script = ""
        if self.waves:
            do_script += "log -recursive /*;"

        if not self.gui:
            do_script += "run -all; quit"

        gpi_if_entry = self.gpi_interfaces[0]
        if gpi_if_entry == "fli":
            lib_opts = [
                "-foreign",
                "cocotb_init "
                + _as_tcl_value(
                    cocotb_tools.config.lib_name_path("fli", "questa").as_posix()
                ),
            ]
        elif gpi_if_entry == "vhpi":
            lib_opts = ["-voptargs=-access=rw+/."]
            lib_opts += [
                "-foreign",
                "vhpi_startup_routines_bootstrap "
                + _as_tcl_value(
                    cocotb_tools.config.lib_name_path("vhpi", "questa").as_posix()
                ),
            ]
        else:
            lib_opts = [
                "-pli",
                _as_tcl_value(
                    cocotb_tools.config.lib_name_path("vpi", "questa").as_posix()
                ),
            ]

        cmds.append(
            ["vsim"]
            + ["-gui" if self.gui else "-c"]
            + ["-onfinish", "stop" if self.gui else "exit"]
            + lib_opts
            + [_as_tcl_value(v) for v in self.test_args]
            + [_as_tcl_value(v) for v in self._get_parameter_options(self.parameters)]
            + [_as_tcl_value(f"{self.hdl_toplevel_library}.{self.sim_hdl_toplevel}")]
            + [_as_tcl_value(v) for v in self.plusargs]
            + pre_cmd
            + ["-do", do_script]
        )

        gpi_extra_list = []
        for gpi_if in self.gpi_interfaces[1:]:
            gpi_if_lib_path = cocotb_tools.config.lib_name_path(gpi_if, "questa")
            if gpi_if_lib_path.is_file():
                gpi_extra_list.append(
                    gpi_if_lib_path.as_posix() + f":cocotb{gpi_if}_entry_point"
                )
            else:
                raise RuntimeError(f"{gpi_if_lib_path} library not found.")
        self.env["GPI_EXTRA"] = ",".join(gpi_extra_list)

        return cmds


class Ghdl(Runner):
    """Implementation of :class:`Runner` for GHDL.

    .. admonition:: Simulator-specific Usage

       * Does not support the ``pre_cmd`` argument to :meth:`.test`.
    """

    supported_gpi_interfaces = {"vhdl": ["vpi"]}

    def _set_env(self) -> None:
        super()._set_env()
        if "COCOTB_TRUST_INERTIAL_WRITES" not in self.env:
            self.env["COCOTB_TRUST_INERTIAL_WRITES"] = "1"

    def _simulator_in_path(self) -> None:
        if shutil.which("ghdl") is None:
            raise SystemExit("ERROR: ghdl executable not found!")

    def _is_mcode_backend(self) -> bool:
        """Is GHDL using the mcode backend?"""
        result = subprocess.run(
            ["ghdl", "--version"],
            check=True,
            universal_newlines=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
        )
        return "mcode" in result.stdout

    def _use_external_viewer(self) -> bool:
        return True

    def _waves_file(self) -> Optional[str]:
        return f"{self.hdl_toplevel}.ghw"

    def _get_include_options(self, includes: Sequence[PathLike]) -> _Command:
        raise RuntimeError

    def _get_define_options(self, defines: Mapping[str, object]) -> _Command:
        raise RuntimeError

    def _get_parameter_options(self, parameters: Mapping[str, object]) -> _Command:
        return [f"-g{name}={value}" for name, value in parameters.items()]

    def _build_command(self) -> List[_Command]:
        sources = self._sources + self._vhdl_sources

        for source in sources:
            if source.tag is not VHDL:
                raise ValueError(
                    f"{type(self).__qualname__} only supports VHDL. {str(source.value)!r} cannot be compiled."
                )

        for arg in self._build_args:
            if arg.tag not in (VHDL, None):
                raise ValueError(
                    f"{type(self).__qualname__} only supports VHDL. build_args {arg.value!r} will not be applied."
                )

        cmds = [
            ["ghdl", "-i"]
            + [f"--work={self.hdl_library}"]
            + [arg.value for arg in self._build_args]
            + [str(source.value) for source in sources]
        ]

        if self.hdl_toplevel is not None:
            cmds += [
                [
                    "ghdl",
                    "-m",
                    f"--work={self.hdl_library}",
                    *(arg.value for arg in self._build_args),
                    self.hdl_toplevel,
                ]
            ]

        return cmds

    def _test_command(self) -> List[_Command]:
        if self.pre_cmd is not None:
            raise RuntimeError("pre_cmd is not implemented for GHDL.")

        ghdl_run_args = self.test_args

        if self._is_mcode_backend() and self.timescale:
            _, precision = self.timescale
            # Convert the time precision to a format string supported by GHDL,
            # if possible.
            # GHDL only supports setting the time precision if the mcode backend
            # is used, using the --time-resolution argument causes GHDL to error
            # out otherwise.
            # https://ghdl.github.io/ghdl/using/InvokingGHDL.html#cmdoption-ghdl-time-resolution
            if precision == "1fs":
                ghdl_time_resolution = "fs"
            elif precision == "1ps":
                ghdl_time_resolution = "ps"
            elif precision == "1ns":
                ghdl_time_resolution = "ns"
            elif precision == "1us":
                ghdl_time_resolution = "us"
            elif precision == "1ms":
                ghdl_time_resolution = "ms"
            elif precision == "1s":
                ghdl_time_resolution = "sec"
            else:
                raise ValueError(
                    "GHDL only supports the following precisions in timescale: 1fs, 1ps, 1us, 1ms, 1s"
                )

            ghdl_run_args.append(f"--time-resolution={ghdl_time_resolution}")

        cmds = [
            ["ghdl", "-r"]
            + [f"--work={self.hdl_toplevel_library}"]
            + ghdl_run_args
            + [self.sim_hdl_toplevel]
            + ["--vpi=" + cocotb_tools.config.lib_name_path("vpi", "ghdl").as_posix()]
            + self.plusargs
            + self._get_parameter_options(self.parameters)
            + ([f"--wave={self._waves_file()}"] if self.waves or self.gui else [])
        ]

        return cmds


class Nvc(Runner):
    """Implementation of :class:`Runner` for NVC.

    .. admonition:: Simulator-specific Usage

       * Does not support the ``pre_cmd`` argument to :meth:`.test`.
       * Does not support the ``timescale`` argument to :meth:`.build` or :meth:`.test`.
    """

    supported_gpi_interfaces = {"vhdl": ["vhpi"]}

    def __init__(self) -> None:
        super().__init__()

        version_str = subprocess.run(
            ["nvc", "--version"],
            check=True,
            universal_newlines=True,
            stdout=subprocess.PIPE,
        ).stdout
        version = NvcVersion.from_commandline(version_str)
        if version > NvcVersion("1.16"):
            self._preserve_case = ["--preserve-case"]
        else:
            self._preserve_case = []

    def _set_env(self) -> None:
        super()._set_env()
        if "COCOTB_TRUST_INERTIAL_WRITES" not in self.env:
            self.env["COCOTB_TRUST_INERTIAL_WRITES"] = "1"

    def _simulator_in_path(self) -> None:
        if shutil.which("nvc") is None:
            raise SystemExit("ERROR: nvc executable not found!")

    def _get_include_options(self, includes: Sequence[PathLike]) -> _Command:
        raise RuntimeError

    def _get_define_options(self, defines: Mapping[str, object]) -> _Command:
        raise RuntimeError

    def _get_parameter_options(self, parameters: Mapping[str, object]) -> _Command:
        return [f"-g{name}={value}" for name, value in parameters.items()]

    def _use_external_viewer(self) -> bool:
        return True

    def _waves_file(self) -> Optional[str]:
        return f"{self.hdl_toplevel}.fst"

    def _build_command(self) -> List[_Command]:
        sources = self._sources + self._vhdl_sources

        for source in sources:
            if source.tag is not VHDL:
                raise ValueError(
                    f"{type(self).__qualname__} only supports VHDL. {str(source.value)!r} cannot be compiled."
                )

        for arg in self._build_args:
            if arg.tag not in (VHDL, None):
                raise ValueError(
                    f"{type(self).__qualname__} only supports VHDL. build_args {arg.value!r} will not be applied."
                )

        cmds = [
            [
                "nvc",
                f"--work={self.hdl_library}",
                "-L",
                str(get_abs_path(self.build_dir)),
            ]
            + [arg.value for arg in self._build_args]
            + ["-a"]
            + [str(source.value) for source in sources]
            + self._preserve_case
        ]

        return cmds

    def _test_command(self) -> List[_Command]:
        work_library = str(get_abs_path(self.build_dir / self.hdl_toplevel_library))
        cmds = [
            [
                "nvc",
                f"--work={self.hdl_toplevel_library}:{work_library}",
                "-L",
                str(get_abs_path(self.build_dir)),
            ]
            + [arg.value for arg in self._build_args]
            + ["-e", self.sim_hdl_toplevel, "--no-save", "--jit"]
            + self.elab_args
            + self._get_parameter_options(self.parameters)
            + ["-r"]
            + self.test_args
            + ["--load=" + cocotb_tools.config.lib_name_path("vhpi", "nvc").as_posix()]
            + self.plusargs
            + ([f"--wave={self._waves_file()}"] if self.waves or self.gui else [])
        ]

        return cmds


class Riviera(Runner):
    """Implementation of :class:`Runner` for Aldec Riviera-PRO.

    .. admonition:: Simulator-specific Usage

       * Does not support the ``pre_cmd`` argument to :meth:`.test`.
       * Does not support the ``gui`` argument to :meth:`.test`.
       * Does not support the ``timescale`` argument to :meth:`.build` or :meth:`.test`.
    """

    supported_gpi_interfaces = {"verilog": ["vpi"], "vhdl": ["vhpi"]}

    def _simulator_in_path(self) -> None:
        if shutil.which("vsimsa") is None:
            raise SystemExit("ERROR: vsimsa executable not found!")

    def _get_include_options(self, includes: Sequence[PathLike]) -> _Command:
        return [f"+incdir+{_as_tcl_value(str(include))}" for include in includes]

    def _get_define_options(self, defines: Mapping[str, object]) -> _Command:
        return [
            f"+define+{name}={self._as_define_value(value)}"
            for name, value in defines.items()
        ]

    def _as_define_value(self, value: object) -> str:
        if isinstance(value, int):
            return str(value)
        elif isinstance(value, str):
            for char in value:
                if ord(char) < 32 or ord(char) >= 255 or char in '\\"':
                    # Control characters are generally not supported.
                    # Not sure if there's any way to escape quotes or backslashes.
                    raise ValueError(
                        f"Character {char!r} not supported in define value"
                    )
            return '\\"\\\\"' + value + '\\\\"\\"'
        else:
            raise TypeError("Can't serialize this type as an SV literal")

    def _get_parameter_options(self, parameters: Mapping[str, object]) -> _Command:
        return [f"-g{name}={value}" for name, value in parameters.items()]

    def _build_command(self) -> List[_Command]:
        do_script: List[str] = ["onerror {\n quit -code 1 \n}"]

        out_file = self.build_dir / self.hdl_library / f"{self.hdl_library}.lib"

        sources = self._sources + self._vhdl_sources + self._verilog_sources

        if outdated(out_file, (source.value for source in sources)) or self.always:
            vhdl_args = [
                arg.value for arg in self._build_args if arg.tag in (VHDL, None)
            ]
            verilog_args = [
                arg.value for arg in self._build_args if arg.tag in (Verilog, None)
            ]
            defines = " ".join(self._get_define_options(self.defines))
            includes = " ".join(self._get_include_options(self.includes))
            verilog_args_str = " ".join(_as_tcl_value(v) for v in verilog_args)
            vhdl_args_str = " ".join(_as_tcl_value(v) for v in vhdl_args)
            hdl_library = _as_tcl_value(self.hdl_library)
            ext_name = _as_tcl_value(
                cocotb_tools.config.lib_name_path("vpi", "riviera").as_posix()
            )

            do_script.append(f"alib {hdl_library}")

            for source in sources:
                if source.tag is Verilog:
                    do_script.append(
                        f"alog -work {hdl_library} -pli {ext_name} -sv {defines} {includes} {verilog_args_str} {_as_tcl_value(str(source.value))}"
                    )
                elif source.tag is VHDL:
                    do_script.append(
                        f"acom -work {hdl_library} {vhdl_args_str} {_as_tcl_value(str(source.value))}"
                    )
                else:
                    raise ValueError(f"Unsupported file type: {source.value}")

        # Explicitly exit the script at the end. In batch mode, which is invoked
        # implicitly by redirecting STDOUT/STDERR of the alog/acom commands,
        # the tool exits by itself even without this 'exit' command -- but not
        # when running from an interactive terminal. Be explicit for predictable
        # behavior.
        do_script.append("exit")

        with tempfile.NamedTemporaryFile(delete=False) as do_file:
            do_file.write("\n".join(do_script).encode())

        return [["vsimsa", "-do", "do", do_file.name]]

    def _test_command(self) -> List[_Command]:
        if self.pre_cmd is not None:
            raise RuntimeError("pre_cmd is not implemented for Riviera.")

        do_script: str = "\nonerror {\n quit -code 1 \n} \n"

        if self.hdl_toplevel_lang == "vhdl":
            do_script += "asim +access +w_nets -interceptcoutput -loadvhpi {EXT_NAME} {EXTRA_ARGS} {TOPLEVEL} {PLUSARGS}\n".format(
                TOPLEVEL=_as_tcl_value(
                    f"{self.hdl_toplevel_library}.{self.sim_hdl_toplevel}"
                ),
                EXT_NAME=_as_tcl_value(
                    cocotb_tools.config.lib_name_path("vhpi", "riviera").as_posix()
                    + ":vhpi_startup_routines_bootstrap"
                ),
                EXTRA_ARGS=" ".join(
                    _as_tcl_value(v)
                    for v in (
                        self.test_args + self._get_parameter_options(self.parameters)
                    )
                ),
                PLUSARGS=" ".join(_as_tcl_value(v) for v in self.plusargs),
            )

            self.env["GPI_EXTRA"] = (
                cocotb_tools.config.lib_name_path("vpi", "riviera").as_posix()
                + ":cocotbvpi_entry_point"
            )
        else:
            do_script += "asim +access +w_nets -interceptcoutput -pli {EXT_NAME} {EXTRA_ARGS} {TOPLEVEL} {PLUSARGS} \n".format(
                TOPLEVEL=_as_tcl_value(
                    f"{self.hdl_toplevel_library}.{self.sim_hdl_toplevel}"
                ),
                EXT_NAME=_as_tcl_value(
                    cocotb_tools.config.lib_name_path("vpi", "riviera").as_posix()
                ),
                EXTRA_ARGS=" ".join(
                    _as_tcl_value(v)
                    for v in (
                        self.test_args + self._get_parameter_options(self.parameters)
                    )
                ),
                PLUSARGS=" ".join(_as_tcl_value(v) for v in self.plusargs),
            )

            self.env["GPI_EXTRA"] = (
                cocotb_tools.config.lib_name_path("vhpi", "riviera").as_posix()
                + ":cocotbvhpi_entry_point"
            )

        if self.waves:
            do_script += "log -recursive /*;"

        do_script += "run -all \nexit"

        with tempfile.NamedTemporaryFile(delete=False) as do_file:
            do_file.write(do_script.encode())

        return [["vsimsa", "-do", "do", do_file.name]]


class Verilator(Runner):
    """Implementation of :class:`Runner` for Verilator.

    .. admonition:: Simulator-specific Usage

       * ``waves=True`` *must* be given to :meth:`.build` if either ``waves`` or ``gui`` are to be used during :meth:`.test`.
       * Does not support the ``pre_cmd`` argument to :meth:`.test`.
    """

    supported_gpi_interfaces = {"verilog": ["vpi"]}

    def _set_env(self) -> None:
        super()._set_env()
        if "COCOTB_TRUST_INERTIAL_WRITES" not in self.env:
            self.env["COCOTB_TRUST_INERTIAL_WRITES"] = "1"

    def _simulator_in_path(self) -> None:
        # the verilator binary is only needed for building
        return

    def _use_external_viewer(self) -> bool:
        return True

    def _waves_file(self) -> Optional[str]:
        return "dump.vcd"

    def _simulator_in_path_build_only(self) -> None:
        executable = shutil.which("verilator")
        if executable is None:
            raise SystemExit("ERROR: verilator executable not found!")
        self.executable: str = executable

    def _get_include_options(self, includes: Sequence[PathLike]) -> _Command:
        return [f"-I{include}" for include in includes]

    def _get_define_options(self, defines: Mapping[str, object]) -> _Command:
        return [f"-D{name}={_as_sv_literal(value)}" for name, value in defines.items()]

    def _get_parameter_options(self, parameters: Mapping[str, object]) -> _Command:
        return [f"-G{name}={value}" for name, value in parameters.items()]

    def _build_command(self) -> List[_Command]:
        self._simulator_in_path_build_only()

        sources = self._sources + self._verilog_sources

        for source in sources:
            if source.tag not in (Verilog, VerilatorControlFile):
                raise ValueError(
                    f"{type(self).__qualname__} only supports Verilog and Verilator Control Files. {str(source.value)!r} cannot be compiled."
                )

        for arg in self._build_args:
            if arg.tag not in (Verilog, None):
                raise ValueError(
                    f"{type(self).__qualname__} only supports Verilog. build_args {arg.value!r} will not be applied."
                )

        if self.hdl_toplevel is None:
            raise ValueError(
                f"{type(self).__qualname__} requires the hdl_toplevel parameter to be specified."
            )

        # TODO: set "--debug" if self.verbose
        # TODO: support "--always"

        verilator_cpp = str(
            cocotb_tools.config.share_dir / "lib" / "verilator" / "verilator.cpp"
        )

        cmds = []
        cmds.append(
            [
                "perl",
                self.executable,
                "-cc",
                "--exe",
                "-Mdir",
                str(self.build_dir),
                "--top-module",
                self.hdl_toplevel,
                "--vpi",
                "--public-flat-rw",
                "--prefix",
                "Vtop",
                "-o",
                self.hdl_toplevel,
                "-LDFLAGS",
                f"-Wl,-rpath,{cocotb_tools.config.libs_dir} -L{cocotb_tools.config.libs_dir} -lcocotbvpi_verilator",
            ]
            + (["--trace"] if self.waves else [])
            + [arg.value for arg in self._build_args]
            + (
                ["--timescale", "{}/{}".format(*self.timescale)]
                if self.timescale is not None
                else []
            )
            + self._get_define_options(self.defines)
            + self._get_include_options(self.includes)
            + self._get_parameter_options(self.parameters)
            + [verilator_cpp]
            + [str(source.value) for source in sources]
        )

        cmds.append(
            [
                "make",
                "-j",
                f"{_get_max_parallel_build_jobs()}",
                "-C",
                str(self.build_dir),
                "-f",
                "Vtop.mk",
                f"VM_TRACE={int(self.waves)}",
            ]
        )

        return cmds

    def _test_command(self) -> List[_Command]:
        if self.pre_cmd is not None:
            raise RuntimeError("pre_cmd is not implemented for Verilator.")

        out_file = self.build_dir / self.sim_hdl_toplevel
        return [
            [str(out_file)]
            + (["--trace"] if self.waves or self.gui else [])
            + self.test_args
            + self.plusargs
        ]


class Xcelium(Runner):
    """Implementation of :class:`Runner` for Cadence Xcelium.

    .. admonition:: Simulator-specific Usage

       * Does not support the ``waves`` argument to :meth:`.build` (must be set in :meth:`.test` instead).
       * Does not support the ``pre_cmd`` argument to :meth:`.test`.
       * Does not support the ``timescale`` argument to :meth:`.test`.
    """

    supported_gpi_interfaces = {"verilog": ["vpi"], "vhdl": ["vhpi"]}

    def _simulator_in_path(self) -> None:
        if shutil.which("xrun") is None:
            raise SystemExit("ERROR: xrun executable not found!")

    def _get_include_options(self, includes: Sequence[PathLike]) -> _Command:
        return [f"-incdir {include}" for include in includes]

    def _get_define_options(self, defines: Mapping[str, object]) -> _Command:
        return [
            f"-define {name}={self._as_define_value(value)}"
            for name, value in defines.items()
        ]

    def _as_define_value(self, value: object) -> str:
        if isinstance(value, int):
            return str(value)
        elif isinstance(value, str):
            for char in value:
                if ord(char) < 32 or ord(char) >= 255 or char == '"':
                    # Control characters are generally not supported.
                    # Not sure if there's any way to escape quotes.
                    raise ValueError(
                        f"Character {char!r} not supported in define value"
                    )
            return '"\\"' + value.replace("\\", "\\\\") + '\\""'
        else:
            raise TypeError("Can't serialize this type as an SV literal")

    def _get_parameter_options(self, parameters: Mapping[str, object]) -> _Command:
        return [f'-gpg "{name} => {value}"' for name, value in parameters.items()]

    def _build_command(self) -> List[_Command]:
        self.env["CDS_AUTO_64BIT"] = "all"

        if self.waves:
            raise RuntimeError(
                "waves is not supported in the build step. Please set it in the test step."
            )

        if self.hdl_toplevel is None:
            raise ValueError("A HDL toplevel is required in all Xcelium compiles.")

        verbosity_opts = []
        if self.verbose:
            verbosity_opts += ["-messages"]
            verbosity_opts += ["-status"]
            verbosity_opts += ["-gverbose"]  # print assigned generics/parameters
            verbosity_opts += ["-pliverbose"]
            verbosity_opts += ["-plidebug"]  # Enhance the profile output with PLI info
            verbosity_opts += [
                "-plierr_verbose"
            ]  # Expand handle info in PLI/VPI/VHPI messages

        else:
            verbosity_opts += ["-quiet"]
            verbosity_opts += ["-plinowarn"]

        sources = self._sources + self._vhdl_sources + self._verilog_sources

        for source in sources:
            if source.tag not in (VHDL, Verilog):
                raise ValueError(f"Unsupported file type: {source.value}")

        vhpi_opts = []
        if any(source.tag is VHDL for source in sources):
            # Xcelium 23.09.004 fixes cocotb issue #1076 as long as the
            # following define is set.
            vhpi_opts.append("-NEW_VHPI_PROPAGATE_DELAY")

        cmds = [
            ["xrun"]
            + ["-logfile"]
            + ["xrun_build.log"]
            + ["-elaborate"]
            + ["-xmlibdirname"]
            + [f"{self.build_dir}/xrun_snapshot"]
            + ["-licqueue"]
            + (["-clean"] if self.always else [])
            + verbosity_opts
            # + ["-vpicompat 1800v2005"]  # <1364v1995|1364v2001|1364v2005|1800v2005> Specify the IEEE VPI
            + ["-access +rwc"]
            + ["-loadvpi"]
            # always start with VPI on Xcelium
            + [
                cocotb_tools.config.lib_name_path("vpi", "xcelium").as_posix()
                + ":vlog_startup_routines_bootstrap"
            ]
            + vhpi_opts
            + [f"-work {self.hdl_library}"]
            + (
                ["-timescale", "{}/{}".format(*self.timescale)]
                if self.timescale is not None
                else []
            )
            + [arg.value for arg in self._build_args]
            + self._get_include_options(self.includes)
            + self._get_define_options(self.defines)
            + self._get_parameter_options(self.parameters)
            + [f"-top {self.hdl_toplevel}"]
            + [str(source_file.value) for source_file in sources]
        ]

        return cmds

    def _test_command(self) -> List[_Command]:
        if self.pre_cmd is not None:
            raise RuntimeError("pre_cmd is not implemented for Xcelium.")

        if self.timescale is not None:
            raise RuntimeError(
                "timescale is not supported in the test step. Please set it in the build step."
            )

        self.env["CDS_AUTO_64BIT"] = "all"

        verbosity_opts = []
        if self.verbose:
            verbosity_opts += ["-messages"]
            verbosity_opts += ["-status"]
            verbosity_opts += ["-gverbose"]  # print assigned generics/parameters
            verbosity_opts += ["-pliverbose"]
            verbosity_opts += ["-plidebug"]  # Enhance the profile output with PLI info
            verbosity_opts += [
                "-plierr_verbose"
            ]  # Expand handle info in PLI/VPI/VHPI messages

        else:
            verbosity_opts += ["-quiet"]
            verbosity_opts += ["-plinowarn"]

        tmpdir = f"implicit_tmpdir_{self.current_test_name}"

        if self.hdl_toplevel_lang == "vhdl":
            xrun_top = ":"
        else:
            xrun_top = self.sim_hdl_toplevel

        if self.waves:
            input_tcl = [
                f'-input "@database -open cocotb_waves -default" '
                f'-input "@probe -database cocotb_waves -create {xrun_top} -all -depth all" '
                f'-input "@run" '
                f'-input "@exit" '
            ]
        else:
            input_tcl = ["-input", "@run; exit;"]

        sources = self._sources + self._vhdl_sources + self._verilog_sources

        vhpi_opts = []
        if any(source.tag is VHDL for source in sources):
            # Xcelium 23.09.004 fixes cocotb issue #1076 as long as the
            # following define is set.
            vhpi_opts.append("-NEW_VHPI_PROPAGATE_DELAY")

        cmds = [["mkdir", "-p", tmpdir]]
        cmds += [
            [
                "xrun",
                "-logfile",
                f"xrun_{self.current_test_name}.log",
                "-xmlibdirname",
                f"{self.build_dir}/xrun_snapshot",
                "-cds_implicit_tmpdir",
                tmpdir,
                "-licqueue",
                *vhpi_opts,
                *verbosity_opts,
                "-R",
                *self.test_args,
                *self.plusargs,
                "-gui" if self.gui else "",
                *input_tcl,
            ]
        ]
        self.env["GPI_EXTRA"] = (
            cocotb_tools.config.lib_name_path("vhpi", "xcelium").as_posix()
            + ":cocotbvhpi_entry_point"
        )

        return cmds


class Vcs(Runner):
    """Implementation of :class:`Runner` for Synopsys VCS.

    .. admonition:: Simulator-specific Usage

       * Does not support the ``pre_cmd`` argument to :meth:`.test`.
       * Does not support VHDL.
       * Does not support the ``timescale`` argument to :meth:`.build` or :meth:`.test`.
    """

    supported_gpi_interfaces = {"verilog": ["vpi"]}

    def _simulator_in_path(self) -> None:
        if shutil.which("vcs") is None:
            raise SystemExit("ERROR: vcs executable not found!")

    def _get_include_options(self, includes: Sequence[PathLike]) -> _Command:
        return [f"+incdir+{include}" for include in includes]

    def _get_define_options(self, defines: Mapping[str, object]) -> _Command:
        return [
            f"+define+{name}={_as_sv_literal(value)}" for name, value in defines.items()
        ]

    def _get_parameter_options(self, parameters: Mapping[str, object]) -> _Command:
        if self.hdl_toplevel is None:
            raise ValueError("A HDL toplevel is required in all VCS compiles.")
        return [
            f"-pvalue+{self.hdl_toplevel}.{name}={value}"
            for name, value in parameters.items()
        ]

    @property
    def sim_file(self) -> Path:
        return self.build_dir / "simv"

    @property
    def _build_opts(self) -> List[str]:
        opts = [
            "-full64",
            "-debug_access+all",
            "+acc+3",
            "-sverilog",
            "-LDFLAGS -Wl,--no-as-needed",
        ]

        if self.verbose:
            opts += ["-diag all"]
        else:
            opts += ["-q"]
            opts += ["-suppress=VPI-CT-NS"]

        return opts

    def _build_command(self) -> List[_Command]:
        cmds: List[_Command] = []
        sources = self._sources + self._vhdl_sources + self._verilog_sources

        for source in sources:
            if source.tag not in (VHDL, Verilog):
                raise ValueError(f"Unsupported file type: {source.value}")

        if outdated(self.sim_file, (source.value for source in sources)) or self.always:
            cmds = [
                ["vcs"]
                + self._build_opts
                + ["-load", cocotb_tools.config.lib_name_path("vpi", "vcs").as_posix()]
                + [arg.value for arg in self._build_args]
                + self._get_include_options(self.includes)
                + self._get_define_options(self.defines)
                + self._get_parameter_options(self.parameters)
                + ["-top", f"{self.hdl_toplevel}"]
                + [str(source.value) for source in sources]
                + ["-o", str(self.sim_file)]
            ]
        else:
            self.log.warning("Skipping compilation of %s", self.sim_file)

        return cmds

    def _test_command(self) -> List[_Command]:
        if self.pre_cmd is not None:
            raise RuntimeError("pre_cmd is not implemented for Vcs.")

        verbosity_opts = []
        if self.verbose:
            verbosity_opts += ["-diag all"]
        else:
            verbosity_opts += ["-suppress=ASLR_DETECTED_INFO"]

        cmds = [[str(self.sim_file), *verbosity_opts, *self.test_args, *self.plusargs]]

        return cmds


class Dsim(Runner):
    """Implementation of :class:`Runner` for Siemens DSim.

    .. admonition:: Simulator-specific Usage

       * Does not support the ``pre_cmd`` argument to :meth:`.test`.
    """

    supported_gpi_interfaces = {"verilog": ["vpi"]}

    def _simulator_in_path(self) -> None:
        if shutil.which("dsim") is None:
            raise SystemExit("ERROR: dsim executable not found!")

    def _get_include_options(self, includes: Sequence[PathLike]) -> _Command:
        return [f"+incdir+{include}" for include in includes]

    def _get_define_options(self, defines: Mapping[str, object]) -> _Command:
        return [
            f"+define+{name}={_as_sv_literal(value)}" for name, value in defines.items()
        ]

    def _get_parameter_options(self, parameters: Mapping[str, object]) -> _Command:
        return [
            f"-defparam {name}={_as_sv_literal(value)}"
            for name, value in parameters.items()
        ]

    @property
    def sim_file(self) -> Path:
        return self.build_dir / "image.so"

    def _use_external_viewer(self) -> bool:
        return True

    def _waves_file(self) -> Optional[str]:
        return "file.vcd"

    def _test_command(self) -> List[_Command]:
        if self.pre_cmd is not None:
            raise RuntimeError("pre_cmd is not implemented for DSim.")

        plusargs = self.plusargs
        if self.waves or self.gui:
            plusargs += [f"-waves {self._waves_file()}"]

        if self.timescale:
            plusargs += ["-timescale {}/{}".format(*self.timescale)]

        return [
            [
                "dsim",
                "-work",
                str(self.build_dir),
                "-pli_lib",
                cocotb_tools.config.lib_name_path("vpi", "dsim").as_posix(),
                "+acc+rwcbfsWF",
                "-image",
                "image",
                *self.test_args,
                *plusargs,
            ]
        ]

    def _build_command(self) -> List[_Command]:
        sources = self._sources + self._verilog_sources

        for source in sources:
            if source.tag is not Verilog:
                raise ValueError(
                    f"{type(self).__qualname__} only supports Verilog. {str(source.value)!r} cannot be compiled."
                )

        for arg in self._build_args:
            if arg.tag not in (Verilog, None):
                raise ValueError(
                    f"{type(self).__qualname__} only supports Verilog. build_args {arg!r} cannot be applied."
                )

        cmds: list[_Command] = []
        if outdated(self.sim_file, (source.value for source in sources)) or self.always:
            cmds = [
                [
                    "dsim",
                    "-work",
                    str(self.build_dir),
                    "-pli_lib",
                    cocotb_tools.config.lib_name_path("vpi", "dsim").as_posix(),
                    "+acc+rwcbfsWF",
                    "-genimage",
                    "image",
                ]
                + self._get_define_options(self.defines)
                + self._get_include_options(self.includes)
                + self._get_parameter_options(self.parameters)
                + [arg.value for arg in self._build_args]
                + [str(source_file.value) for source_file in sources]
            ]

        else:
            self.log.warning("Skipping compilation of %s", self.sim_file)

        return cmds


def get_runner(simulator_name: str) -> Runner:
    """Return an instance of a runner for *simulator_name*.

    Args:
        simulator_name: Name of simulator to get runner for.

    Raises:
        ValueError: If *simulator_name* is not one of the supported simulators or an alias of one.
    """

    supported_sims: Dict[str, Type[Runner]] = {
        "icarus": Icarus,
        "questa": Questa,
        "ghdl": Ghdl,
        "riviera": Riviera,
        "verilator": Verilator,
        "xcelium": Xcelium,
        "nvc": Nvc,
        "vcs": Vcs,
        "dsim": Dsim,
        # TODO: "activehdl": ActiveHdl,
    }
    try:
        return supported_sims[simulator_name]()
    except KeyError:
        raise ValueError(
            f"Simulator {simulator_name!r} is not in supported list: {', '.join(supported_sims)}"
        ) from None
